from numpy import ones
from numpy import zeros
from numpy import array
from numpy import linspace

from compas_tno.shapes import rectangular_topology
from compas_tno.shapes import MeshDos

import math


def pointed_arch_shape(hc=1.00, L=2.0, x0=0.0, thk=0.20, b=0.5, t=5.0, total_nodes=100):
    """Function to create meshes to define upper and lower bounds of 2D pointed arch.

    Parameters
    ----------
    hc : float, optional
        Rise of the arch measured with regards to the center line, by default 1.00
    L : float, optional
        Span of the arch considered as center, to center. (L <= 2*H), by default 2.0
    x0 : float, optional
        Beginning of the linear form diagram, by default 0.0
    thk : float, optional
        Thickness of the arch, by default 0.20
    b : float, optional
        Out of plane dimension of the arch, by default 0.5
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 5.0
    total_nodes : int, optional
        Numbers of nodes to be considered in the form diagram, by default 100

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape
    """

    radius = 1/L * (hc**2 + L**2/4)
    ri = radius - thk/2
    re = radius + thk/2
    # print('radius/ ri / re =', radius, ri, re)
    zc = 0.0
    xc1 = x0 + radius
    xc2 = x0 + L - radius
    # print('Centers x =', xc1, xc2)

    x = linspace(x0, x0 + L, num=total_nodes, endpoint=True)
    xs, ys, faces_i = rectangular_topology(x, [-b/2, 0, b/2])

    i = 0

    zts = []
    zis = []
    zes = []

    for xi in x:
        if xi <= x0 + L/2:
            dx = xi - xc1
        else:
            dx = xc2 - xi

        zi2 = radius**2 - dx**2
        if zi2 > 0:
            zi = math.sqrt(zi2) - zc
        else:
            zi = 0
        zei = math.sqrt(re**2 - dx**2) + zc
        zii2 = ri**2 - dx**2
        if zii2 > 0:
            zii = math.sqrt(zii2) - zc
        else:
            zii = -t

        zts += [zi, zi, zi]
        zis += [zii, zii, zii]
        zes += [zei, zei, zei]

        i = i + 1

    xyzlb = array([xs, ys, zis]).transpose()
    xyzub = array([xs, ys, zes]).transpose()
    xyzt = array([xs, ys, zts]).transpose()

    intrados = MeshDos.from_vertices_and_faces(xyzlb, faces_i)
    extrados = MeshDos.from_vertices_and_faces(xyzub, faces_i)
    middle = MeshDos.from_vertices_and_faces(xyzt, faces_i)

    return intrados, extrados, middle


def pointed_arch_ub_lb_update(x, y, thk, t, hc, L, x0):
    """Update upper and lower bounds of a pointed arch based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    hc : float
        Rise of the arch measured with regards to the center line
    L : float
        Span of the arch considered as center, to center. (L <= 2*H)
    x0 : float
        Beginning of the linear form diagram

    Returns
    -------
    ub : array
        Values of the upper bound in the points
    lb : array
        Values of the lower bound in the points
    """

    radius = 1/L * (hc**2 + L**2/4)
    ri = radius - thk/2
    re = radius + thk/2
    zc = 0.0
    xc1 = x0 + radius
    xc2 = x0 + L - radius

    ub = ones((len(x), 1))
    lb = ones((len(x), 1)) * - t

    for i in range(len(x)):
        if x[i] <= x0 + L/2:
            dx = x[i] - xc1
        else:
            dx = xc2 - x[i]
        ub[i] = math.sqrt(re**2 - dx**2) - zc
        lb2 = ri**2 - dx**2
        if lb2 > 0:
            lb[i] = math.sqrt(lb2) - zc

    return ub, lb


def pointed_arch_dub_dlb(x, y, thk, t, hc, L, x0):
    """Computes the sensitivities of upper and lower bounds of a pointed arch based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    hc : float
        Rise of the arch measured with regards to the center line
    L : float
        Span of the arch considered as center, to center. (L <= 2*H)
    x0 : float
        Beginning of the linear form diagram

    Returns
    -------
    dub : array
        Values of the sensitivities for the upper bound in the points
    dlb : array
        Values of the sensitivities for the lower bound in the points
    """

    radius = 1/L * (hc**2 + L**2/4)
    ri = radius - thk/2
    re = radius + thk/2
    xc1 = x0 + radius
    xc2 = x0 + L - radius
    zc = 0.0

    dub = zeros((len(x), 1))
    dlb = zeros((len(x), 1))

    for i in range(len(x)):
        if x[i] <= x0 + L/2:
            dx = x[i] - xc1
        else:
            dx = xc2 - x[i]
        ze = math.sqrt(re**2 - dx**2) - zc
        dub[i] = re/(2*ze)
        zi2 = ri**2 - dx**2
        if zi2 > 0:
            zi = math.sqrt(zi2) - zc
            dlb[i] = - ri/(2*zi)

    return dub, dlb  # ub, lb


# def pointed_arch_b_update(x, y, thk, fixed, H=1.00, L=2.0, x0=0.0):

#     radius = H / 2 + (L**2 / (8 * H))
#     re = radius + thk/2
#     zc = radius - H
#     b = zeros((len(fixed), 2))
#     x_lim = math.sqrt(re**2 - zc**2)

#     for i in range(len(fixed)):
#         b[i, :] = [x_lim - L/2, 0.0]

#     return b


# def pointed_arch_db(x, y, thk, fixed, H=1.00, L=2.0, x0=0.0):  # This does not work for spring angles

#     radius = H / 2 + (L**2 / (8 * H))
#     re = radius + thk/2
#     zc = radius - H
#     db = zeros((len(fixed), 2))
#     x_lim = math.sqrt(re**2 - zc**2)

#     for i in range(len(fixed)):
#         db[i, :] = [1/2 * re/x_lim, 0.0]

#     return db



"""
********************************************************************************
compas_tno.analysis
********************************************************************************

.. currentmodule:: compas_tno.analysis

Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Analysis

Routines
========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    limit_analysis_GSF
    thk_minmax_GSF
    max_n_minmax_GSF

"""

from .analysis import (
    Analysis
)

from .routines import (
    limit_analysis_GSF,
    thk_minmax_GSF,
    max_n_minmax_GSF
)

__all__ = [
    'Analysis',

    'limit_analysis_GSF',
    'thk_minmax_GSF',
    'max_n_minmax_GSF'
]




from compas_tno.shapes.dome import dome_zt_update
from compas_tno.shapes.crossvault import crossvault_middle_update
from compas_tno.shapes.pointed_crossvault import pointed_vault_middle_update
from compas_tno.utilities.interpolation import get_shape_middle_pattern


def apply_selfweight_from_shape_proxy(formdata, shapedata):
    """Apply selfweight to the nodes of the form diagram based on the shape from proxy

    Parameters
    ----------
    formdata : dict
        Data of the form diagram to apply the selfweight.
    shapedata : Shape
        Data of the shape of the masonry.

    Returns
    -------
    formdata
        Data of the form diagram with loads applied
    """

    from compas_tno.diagrams import FormDiagram
    from compas_tno.shapes import Shape

    form = FormDiagram.from_data(formdata)
    shape = Shape.from_data(shapedata)

    apply_selfweight_from_shape(form, shape)

    return form.to_data()


def apply_selfweight_from_shape(form, shape, pz_negative=True, normalize=True):
    """Apply selfweight to the nodes of the form diagram based on the shape

    Parameters
    ----------
    form : FormDiagram
        Form diagram to apply the selfweight
    shape : Shape
        Shape of the masonry
    pz_negative : bool, optional
        Wether or not the vertical loads are negative, by default True
    normalize : bool, optional
        Wether or not normalise the selfweight to match shape.total_weight, by default True

    Returns
    -------
    None
        The FormDiagram is modified in place
    """

    form_ = form.copy()
    total_selfweight = shape.compute_selfweight()
    ro = shape.ro
    thk = shape.datashape['thk']

    x = form.vertices_attribute('x')
    y = form.vertices_attribute('y')

    if shape.datashape['type'] == 'dome':
        zt = dome_zt_update(x, y, shape.datashape['radius'], shape.datashape['t'], shape.datashape['center'])
    elif shape.datashape['type'] == 'crossvault':
        zt = crossvault_middle_update(x, y,  shape.datashape['t'],  xy_span=shape.datashape['xy_span'])
    elif shape.datashape['type'] == 'pointed_crossvault':
        zt = pointed_vault_middle_update(x, y,  shape.datashape['t'],  xy_span=shape.datashape['xy_span'],
                                         hc=shape.datashape['hc'], he=shape.datashape['he'], hm=shape.datashape['hm'])
    else:
        XY = form.vertices_attributes('xy')
        zt = get_shape_middle_pattern(shape, XY)

    i = 0
    for key in form_.vertices():
        z = float(zt[i])
        form_.vertex_attribute(key, 'z', value=z)
        form.vertex_attribute(key, 'target', value=z)
        i += 1

    pzt = 0
    for key in form.vertices():
        pz = form_.vertex_area(key)
        form.vertex_attribute(key, 'pz', value=pz)
        pzt += pz

    if shape.datashape['type'] == 'arch' or shape.datashape['type'] == 'pointed_arch':
        pzt = 0
        for key in form.vertices():
            form.vertex_attribute(key, 'pz', value=1.0)
            if form.vertex_attribute(key, 'is_fixed') is True:
                form.vertex_attribute(key, 'pz', value=0.5)
            pzt += form.vertex_attribute(key, 'pz')

    factor = 1.0 * ro * thk  # Transform tributary area in tributary load
    if normalize:
        factor = total_selfweight/pzt
    if pz_negative:
        factor *= -1  # make loads negative

    for key in form.vertices():
        pzi = factor * form.vertex_attribute(key, 'pz')
        form.vertex_attribute(key, 'pz', value=pzi)


def apply_selfweight_from_pattern(form, pattern, plot=False, pz_negative=True, tol=10e-4):
    """Apply selfweight to the nodes considering a different Form Diagram to locate loads.
    Note: the base pattern has to coincide with nodes from the original form diagram.

    Parameters
    ----------
    form : FormDiagram
        Form diagram to apply the selfweight.
    pattern : Mesh
        Mesh in which the forces should be based.
    plot : bool, optional
        Whether or not plot the diagrams and its overlap, by False
    pz_negative : bool, optional
        Wether or not the vertical loads are negative by default ``True``.
    tol : float, optional
        Tolerance for the nodal position match of form and pattern, by default ``10e-4``.

    Returns
    -------
    None
        The FormDiagram is modified in place.
    """

    form_ = pattern

    form.vertices_attribute('pz', 0.0)
    key_real_to_key = {}

    for key in form_.vertices():
        x, y, _ = form_.vertex_coordinates(key)
        for key_real in form.vertices():
            x_real, y_real, _ = form.vertex_coordinates(key_real)
            if x - tol < x_real < x + tol and y - tol < y_real < y + tol:
                key_real_to_key[key_real] = key
                break

    pzt = 0
    for key in key_real_to_key:
        pz = form_.vertex_attribute(key_real_to_key[key], 'pz')
        # if pz_negative:
        #     pz *= -1  # make loads negative
        form.vertex_attribute(key, 'pz', value=pz)
        pzt += pz

    if plot:
        print('total load applied:', pzt)


def apply_selfweight_from_thrust(form, thickness=0.5, density=20.0):
    """Lump the selfweight in the nodes of the thrust network based on their current position.
    The loads are computed based on the tributary area times the thickness times the density.
    For variable thickness the nodal attribute `thk` is considered.

    Parameters
    ----------
    form : FormDiagram
        The form diagram to be considered
    thickness : float, optional
        The thickness of the problem, by default 0.50
        If None is passed, the thickness is taken from the nodal attribute `thk`
    density : float, optional
        The density of the material, by default 20.0

    Return
    ------
    None
        The form diagram is updated in place
    """

    for key in form.vertices():
        ai = form.vertex_area(key)
        if thickness:
            load = -1 * ai * thickness * density
        else:
            thk = form.vertex_attribute(key, 'thk')
            load = -1 * ai * thk * density
        form.vertex_attribute(key, 'pz', load)


def apply_horizontal_multiplier(form, lambd=1.0, direction='x'):
    """Modify the applied loads considering a load multiplier.

    Parameters
    ----------
    form : FormDiagram
        Form diagram to apply the horizontal multiplier.
    lambd : float, optional
        Value of the horizontal multiplier, by default ``1.0``.
    direction : str, optional
        Direction to apply the loads, ``x`` or ``y``, by default ``x``.

    Returns
    -------
    None
        The FormDiagram is modified in place.
    """

    arg = 'p' + direction

    for key in form.vertices():
        pz = form.vertex_attribute(key, 'pz')
        form.vertex_attribute(key, arg, -1 * pz * lambd)  # considers that swt (pz) is negative


def apply_fill_load(form):
    """Modify the applied loads considering a fill.

    Parameters
    ----------
    form : FormDiagram
        Form diagram to apply the horizontal multiplier.

    Note
    -------
        In development.
    """

    raise NotImplementedError('Not implemented')



from compas_tno.diagrams import FormDiagram
from compas_tno.shapes import Shape
from compas_tno.analysis import Analysis
from compas_tno.viewers import Viewer

form = FormDiagram.create_cross_form()
shape = Shape.create_crossvault(spr_angle=30.0)

analysis = Analysis.create_lp_analysis(form, shape, solver='CVXPY', printout=True)
analysis.apply_selfweight()
analysis.set_up_optimiser()
analysis.run()

view = Viewer(form)
view.draw_thrust()
view.show()



from compas_tno.shapes import Shape
from compas_tno.diagrams import FormDiagram
from compas_tno.viewers import Viewer
from compas_tno.analysis import Analysis
from compas.geometry import Vector, Point
from numpy import array


# ----------------------------------------
# 1. Shape geometric definition
# ----------------------------------------
radius = 5.0
thk = 0.50
center = [5, 5]
dome = Shape.create_dome(radius=radius, thk=thk, center=center)
dome.ro = 1.0

# ----------------------------------------
# 2. Form diagram geometric definition
# ----------------------------------------
discretisation = [16, 20]
form = FormDiagram.create_circular_radial_form(radius=radius, discretisation=discretisation)

# ----------------------------------------
# 3. Define displacement field
# ----------------------------------------

vector_supports = []
vectors_plot = []
base_plot = []
xc = center[0]

for key in form.vertices_where({'is_fixed': True}):
    x, y, z = form.vertex_coordinates(key)
    dXbi = [0, 0, 0]
    if x - xc > 0.1:
        dXbi = [1, 0, 0]
        vectors_plot.append(Vector(*dXbi))
        base_plot.append(Point(x, y, z))
    if x - xc < -0.1:
        dXbi = [-1, 0, 0]
        vectors_plot.append(Vector(*dXbi))
        base_plot.append(Point(x, y, z))

    vector_supports.append(dXbi)

dXb = array(vector_supports)

# --------------------------------------------
# 4. Create analysis, run and visualise
# --------------------------------------------
analysis = Analysis.create_compl_energy_analysis(form, dome, solver='IPOPT', support_displacement=dXb, printout=True)
analysis.optimiser.set_constraints(['funicular', 'envelope', 'reac_bounds'])
analysis.apply_selfweight()
analysis.apply_envelope()
analysis.apply_reaction_bounds()
analysis.set_up_optimiser()
analysis.run()

view = Viewer(form, dome)
view.scale_edge_thickness(5.0)
view.draw_form()
view.draw_shape()
view.draw_reactions(extend_reactions=True)
view.draw_cracks()
for i in range(len(vectors_plot)):
    vector = vectors_plot[i]
    base = base_plot[i]
    view.draw_vector(vector=vector, base=base)
view.show()



import math


__all__ = [
    'paraboloid',
    'dome'
]


def paraboloid(x, y, coef=-0.1):
    """Function of a paraboloid

    Parameters
    ----------
    x : float
        x-coordinate
    y : float
        y-coordinate
    coef : float, optional
        Coefficient of the equation, by default -0.1

    Returns
    -------
    z : float
        Value at (x, y)
    """

    return coef * (x*x + y*y)


def dome(x, y, r):
    """Function of a dome

    Parameters
    ----------
    x : float
        x-coordinate
    y : float
        y-coordinate
    r : float
        Radius

    Returns
    -------
    z : float
        Value at (x, y)
    """

    return math.sqrt(r*r - x*x - y*y)



from compas_tno.shapes import arch_ub_lb_update
from compas_tno.shapes import arch_b_update
from compas_tno.shapes import arch_dub_dlb
from compas_tno.shapes import arch_db

from compas_tno.shapes import pointed_arch_ub_lb_update
from compas_tno.shapes import pointed_arch_dub_dlb

from compas_tno.shapes import dome_ub_lb_update
from compas_tno.shapes import dome_b_update
from compas_tno.shapes import dome_dub_dlb
from compas_tno.shapes import dome_db
from compas_tno.shapes import dome_b_update_with_n
from compas_tno.shapes import dome_db_with_n

from compas_tno.shapes import pavillionvault_ub_lb_update
from compas_tno.shapes import pavillionvault_b_update
from compas_tno.shapes import pavillionvault_dub_dlb
from compas_tno.shapes import pavillionvault_db

from compas_tno.shapes import crossvault_ub_lb_update
from compas_tno.shapes import crossvault_dub_dlb

from compas_tno.shapes import pointed_vault_ub_lb_update
from compas_tno.shapes import pointed_vault_dub_dlb

from compas_tno.shapes import general_ub_lb_update_with_s
from compas_tno.shapes import general_dub_dlb_with_s
from compas_tno.shapes import general_ub_lb_update_with_n
from compas_tno.shapes import general_dub_dlb_with_n
# from compas_tno.shapes import general_b_update_with_n
# from compas_tno.shapes import general_db_with_n
from compas_tno.shapes import general_ub_lb_update_with_t_middle_constant
from compas_tno.shapes import general_db_with_t_middle_constant
from compas_tno.shapes import general_ub_lb_update_with_t_middle_variable
from compas_tno.shapes import general_db_with_t_middle_variable
from compas_tno.shapes import general_ub_lb_update_with_t_intrados
from compas_tno.shapes import general_db_with_t_intrados


def ub_lb_update(x, y, thk, t, shape, ub, lb, s, variables):
    """Function to update the ub-lb vertical bounds of the vertices.

    Parameters
    ----------
    x : array
        xcoordinates of the pattern
    y : array
        y-coordinates of the pattern
    thk : float
        thickness to update the bounds to
    t : float
        thickness considered to the outer vertices
    shape : Shape
        Shape object to be updated
    ub : array
        Current upper-bound limits
    lb : array
        Current lower-bound limits
    s : aray
        Original level of the supports
    variables : list
        List with the variables passed

    Returns
    -------
    ub, lb: array
        New position of the bounds in the point analysed
    """

    if shape.datashape['type'] == 'arch':
        return arch_ub_lb_update(x, y, thk, t, H=shape.datashape['H'], L=shape.datashape['L'], x0=shape.datashape['x0'])
    if shape.datashape['type'] == 'pointed_arch':
        return pointed_arch_ub_lb_update(x, y, thk, t, hc=shape.datashape['hc'], L=shape.datashape['L'], x0=shape.datashape['x0'])
    elif shape.datashape['type'] == 'dome':
        return dome_ub_lb_update(x, y, thk, t, center=shape.datashape['center'], radius=shape.datashape['radius'])
    elif shape.datashape['type'] == 'crossvault':
        return crossvault_ub_lb_update(x, y, thk, t, xy_span=shape.datashape['xy_span'])
    elif shape.datashape['type'] == 'pavillionvault':
        return pavillionvault_ub_lb_update(x, y, thk, t, xy_span=shape.datashape['xy_span'])
    elif shape.datashape['type'] == 'pointed_crossvault':
        return pointed_vault_ub_lb_update(x, y, thk, t, xy_span=shape.datashape['xy_span'], hc=shape.datashape['hc'], he=shape.datashape['he'], hm=shape.datashape['hm'])
    elif shape.datashape['type'] == 'general':
        if 't' in variables:
            thickness_type = shape.datashape['thickness_type']
            if thickness_type == 'constant':
                return general_ub_lb_update_with_t_middle_constant(thk, s, shape.middle, t)  # here open up something
            elif thickness_type == 'variable':
                return general_ub_lb_update_with_t_middle_variable(thk, s, shape.middle, t)  # here open up something
            elif thickness_type == 'intrados':
                return general_ub_lb_update_with_t_intrados(thk, lb, shape.intrados, t)  # here open up something
            else:
                raise Exception
        elif 's' in variables:
            return general_ub_lb_update_with_s(ub, lb, thk)  # thk is 's' in this equation
        elif 'n' in variables:
            return general_ub_lb_update_with_n(ub, lb, thk, shape.intrados, shape.extrados, t)  # thk is 'n' in this equation
        else:
            raise Exception
    else:
        raise Exception


def dub_dlb_update(x, y, thk, t, shape, ub, lb, s, variables):
    """Function to update the derivatives of the ub-lb vertical bounds of the vertices.

    Parameters
    ----------
    x : array
        xcoordinates of the pattern
    y : array
        y-coordinates of the pattern
    thk : float
        thickness to update the bounds to
    t : float
        thickness considered to the outer vertices
    shape : Shape
        Shape object to be updated
    ub : array
        Current upper-bound limits
    lb : array
        Current lower-bound limits
    s : aray
        Original level of the supports
    variables : list
        List with the variables passed

    Returns
    -------
    dub, dlb
        New sensitivities of the bounds in the point analysed
    """

    if shape.datashape['type'] == 'arch':
        return arch_dub_dlb(x, y, thk, t, H=shape.datashape['H'], L=shape.datashape['L'], x0=shape.datashape['x0'])
    if shape.datashape['type'] == 'pointed_arch':
        return pointed_arch_dub_dlb(x, y, thk, t, hc=shape.datashape['hc'], L=shape.datashape['L'], x0=shape.datashape['x0'])
    elif shape.datashape['type'] == 'dome':
        return dome_dub_dlb(x, y, thk, t, center=shape.datashape['center'], radius=shape.datashape['radius'])
    elif shape.datashape['type'] == 'crossvault':
        return crossvault_dub_dlb(x, y, thk, t, xy_span=shape.datashape['xy_span'])
    elif shape.datashape['type'] == 'pavillionvault':
        return pavillionvault_dub_dlb(x, y, thk, t, xy_span=shape.datashape['xy_span'])
    elif shape.datashape['type'] == 'pointed_crossvault':
        return pointed_vault_dub_dlb(x, y, thk, t, xy_span=shape.datashape['xy_span'], hc=shape.datashape['hc'], he=shape.datashape['he'], hm=shape.datashape['hm'])
    elif shape.datashape['type'] == 'general':
        if 't' in variables:
            thickness_type = shape.datashape['thickness_type']
            if thickness_type == 'constant':
                return general_db_with_t_middle_constant(s, shape.middle)  # here open up something
            elif thickness_type == 'variable':
                return general_db_with_t_middle_variable(s, shape.middle)  # here open up something
            elif thickness_type == 'intrados':
                return general_db_with_t_intrados(lb, shape.intrados)  # here open up something
            else:
                raise Exception
        elif 's' in variables:
            return general_dub_dlb_with_s(ub, lb)  # thk is 's' in this equation
        elif 'n' in variables:
            return general_dub_dlb_with_n(ub, lb, thk, shape.intrados, shape.extrados, t)  # thk is 'n' in this equation
        else:
            raise Exception
    else:
        raise Exception


def b_update(x, y, thk, fixed, shape, b, variables):
    """Function to update the limits of the extension of the reaction forces on the support vertices.

    Parameters
    ----------
    x : array
        xcoordinates of the pattern
    y : array
        y-coordinates of the pattern
    thk : float
        thickness to update the bounds to
    fixed : list
        LIst with the indices of the fixed vertices
    shape : Shape
        Shape object to be updated
    b : array
        Current ``b`` limits

    Returns
    -------
    b : list
        New ``b`` limits
    """

    if shape.datashape['type'] == 'arch':
        return arch_b_update(x, y, thk, fixed, H=shape.datashape['H'], L=shape.datashape['L'], x0=shape.datashape['x0'])
    elif shape.datashape['type'] == 'dome':
        return dome_b_update(x, y, thk, fixed, center=shape.datashape['center'], radius=shape.datashape['radius'])
    elif shape.datashape['type'] == 'pavillionvault':
        return pavillionvault_b_update(x, y, thk, fixed, xy_span=shape.datashape['xy_span'])
    elif shape.datashape['type'] == 'general':
        if 'n' in variables:
            if shape.datashape['base_structure']['type'] == 'dome' or shape.datashape['base_structure']['type'] == 'dome_polar':
                return dome_b_update_with_n(x, y, thk, fixed, b, center=shape.datashape['base_structure']['center'])
            else:
                raise Exception
                # return general_b_update_with_n(b, thk, fixed)  # thk is 'n' in this equation
        else:
            raise Exception
    else:
        raise Exception


def db_update(x, y, thk, fixed, shape, b, variables):
    """Function to update the derrivatives of the limits of the extension of the reaction forces on the support vertices.

    Parameters
    ----------
    x : array
        xcoordinates of the pattern
    y : array
        y-coordinates of the pattern
    thk : float
        thickness to update the bounds to
    fixed : list
        LIst with the indices of the fixed vertices
    shape : Shape
        Shape object to be updated
    b : array
        Current ``b`` limits

    Returns
    -------
    db
        Sensitivities of the ``b`` limits in the point
    """

    if shape.datashape['type'] == 'arch':
        return arch_db(x, y, thk, fixed, H=shape.datashape['H'], L=shape.datashape['L'], x0=shape.datashape['x0'])
    elif shape.datashape['type'] == 'dome':
        return dome_db(x, y, thk, fixed, center=shape.datashape['center'], radius=shape.datashape['radius'])
    elif shape.datashape['type'] == 'pavillionvault':
        return pavillionvault_db(x, y, thk, fixed, xy_span=shape.datashape['xy_span'])
    elif shape.datashape['type'] == 'general':
        if 'n' in variables:
            if shape.datashape['base_structure']['type'] == 'dome':
                return dome_db_with_n(x, y, fixed, center=shape.datashape['base_structure']['center'])
            else:
                raise Exception
                # return general_b_update_with_n(b, thk, fixed)  # thk is 'n' in this equation
        else:
            raise Exception
    else:
        raise Exception



"""
********************************************************************************
compas_tno
********************************************************************************

.. currentmodule:: compas_tno


.. toctree::
    :maxdepth: 1

    compas_tno.algorithms
    compas_tno.analysis
    compas_tno.diagrams
    compas_tno.optimisers
    compas_tno.plotters
    compas_tno.problems
    compas_tno.shapes
    compas_tno.solvers
    compas_tno.utilities
    compas_tno.viewers

"""

from __future__ import print_function

import os

__author__ = ["Ricardo Maia Avelino"]
__copyright__ = "Block Research Group - ETH Zurich"
__license__ = "MIT License"
__email__ = "maia@arch.ethz.ch"
__version__ = "0.2.1"


HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))
TEMP = os.path.abspath(os.path.join(HOME, "temp"))


__all__ = ["HOME", "DATA", "DOCS", "TEMP"]


def get(filename=''):
    """Get the full path to one of the sample data files.
    Parameters
    ----------
    filename : str
        The name of the data file.
        The following are available.
        *
    Returns
    -------
    str
        The full path to the specified file.
    Notes
    -----
    The file name should be specified relative to the ``compas_tno`` sample data folder.
    This folder is only locally available if you installed ``compas_tno`` from source,
    or if you are working directly with the source.
    In all other cases, the function will get the corresponding files direcly from
    the GitHub repo, at https://raw.githubusercontent.com/BlockResearchGroup/compas_tno/master/data.
    Examples
    --------
    The ``compas_tno.get`` function is meant to be used in combination with the static
    constructors of the data structures.
    .. code-block:: python
        import compas_tno
        from compas_tno.diagrams import FormDiagram
        form = FormDiagram.from_obj(compas.get('faces.obj'))
    """
    filename = filename.strip('/')

    localpath = os.path.abspath(os.path.join(DATA, filename))

    return localpath

    # if os.path.exists(localpath):
    #     return localpath
    # else:
    #     return "https://raw.githubusercontent.com/BlockResearchGroup/compas_tno/master/data/{}".format(filename)


def wakeup_matlab():
    """ Method to wake up matlab in the background. Initialize MATLAB. start_matlab

    Note: This function has to be called directly from the terminal.
    """

    import matlab.engine
    future = matlab.engine.connect_matlab(background=True)
    eng = future.result()
    eng.cvx_begin(nargout=0)
    print(matlab.engine.find_matlab())

    return eng


def open_dialog():
    """Method to open a dialog window to find the path of a file.

    Returns
    -------
    file_path
        The path of the file selected
    """

    import tkinter as tk
    from tkinter import filedialog

    root = tk.Tk()
    root.withdraw()

    file_path = filedialog.askopenfilename()

    return file_path



import math
from compas.utilities import geometric_key
from compas.geometry import mirror_points_line
from compas.geometry import rotate_points_xy
from compas.datastructures import Mesh

from compas.geometry import distance_point_point_xy
from compas.datastructures import mesh_weld


def create_cross_form(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=10, fix='corners'):
    """Construct a FormDiagram based on cross discretiastion with orthogonal arrangement and diagonal.

    Parameters
    ----------
    xy_span : list, optional
        List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted., by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.

    Notes
    ----------------------
    Position of the quadrants is as in the schema below:

        Q3
    Q2      Q1
        Q4
    """

    if isinstance(discretisation, list):
        discretisation = discretisation[0]

    y1 = float(xy_span[1][1])
    y0 = float(xy_span[1][0])
    x1 = float(xy_span[0][1])
    x0 = float(xy_span[0][0])
    x_span = x1 - x0
    y_span = y1 - y0
    dx = x_span/discretisation
    dy = y_span/discretisation

    lines = []

    for i in range(discretisation+1):
        for j in range(discretisation+1):
            if i < discretisation and j < discretisation:
                # Vertical Members:
                xa = x0 + dx*i
                ya = y0 + dy*j
                xb = x0 + dx*(i + 1)
                yb = y0 + dy*j
                # Horizontal Members:
                xc = x0 + dx*i
                yc = y0 + dy*j
                xd = x0 + dx*i
                yd = y0 + dy*(j + 1)
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
                lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                if i == j:
                    # Diagonal Members in + Direction:
                    xc = x0 + dx*i
                    yc = y0 + dy*j
                    xd = x0 + dx*(i + 1)
                    yd = y0 + dy*(j + 1)
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                if i + j == discretisation:
                    # Diagonal Members in - Direction:
                    xc = x0 + dx*i
                    yc = y0 + dy*j
                    xd = x0 + dx*(i - 1)
                    yd = y0 + dy*(j + 1)
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                    if i == (discretisation - 1):
                        xc = x0 + dx*i
                        yc = y0 + dy*j
                        xd = x0 + dx*(i + 1)
                        yd = y0 + dy*(j - 1)
                        lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
            else:
                if i == discretisation and j < discretisation:
                    # Vertical Members on last column:
                    xa = x0 + dx*j
                    ya = y0 + dy*i
                    xb = x0 + dx*(j + 1)
                    yb = y0 + dy*i
                    # Horizontal Members:
                    xc = x0 + dx*i
                    yc = y0 + dy*j
                    xd = x0 + dx*i
                    yd = y0 + dy*(j + 1)
                    lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])

    mesh = Mesh.from_lines(lines, delete_boundary_face=True)
    form = cls.from_mesh(mesh)
    # form = cls.from_lines(lines, delete_boundary_face=True)
    gkey_key = form.gkey_key()

    if fix == 'corners':
        form.vertex_attribute(gkey_key[geometric_key([x0, y0, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x0, y1, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x1, y0, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x1, y1, 0.0])], 'is_fixed', True)
    else:
        [bnds] = form.vertices_on_boundaries()
        for key in bnds:
            form.vertex_attribute(key, 'is_fixed', True)
        form = form.delete_boundary_edges()  # Check if this should be here, or explicit

    return form


def create_cross_diagonal(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], partial_bracing_modules=None, discretisation=10, fix='corners'):
    """Construct a FormDiagram based on a mixture of cross and fan discretiastion

    Parameters
    ----------
    xy_span : list, optional
        List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    partial_bracing_modules : str, optional
        If partial bracing modules are included, by default None
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.
    """

    if isinstance(discretisation, list):
        discretisation = discretisation[0]

    y1 = float(xy_span[1][1])
    y0 = float(xy_span[1][0])
    x1 = float(xy_span[0][1])
    x0 = float(xy_span[0][0])
    x_span = x1 - x0
    y_span = y1 - y0
    dx = x_span/discretisation
    dy = y_span/discretisation

    xc0 = x0 + x_span/2
    yc0 = y0 + y_span/2

    nx = ny = int(discretisation/2)
    if partial_bracing_modules is None:
        nstop = 0
    else:
        nstop = nx - partial_bracing_modules  # Test to stop

    line_hor = [[x0, yc0, 0.0], [xc0, yc0, 0.0]]
    line_ver = [[xc0, y0, 0.0], [xc0, yc0, 0.0]]
    origin = [xc0, yc0, 0.0]

    lines = []

    for i in range(nx):
        for j in range(ny+1):
            if j <= i:
                if i >= nstop and j >= nstop:
                    # Diagonal Members:
                    xa = x0 + dx * i
                    ya = y0 + dy * 1 * (i - j) / (nx - j) + dy * j
                    xb = x0 + dx * (i + 1)
                    yb = y0 + dy * 1 * (i - j + 1) / (nx - j) + dy * j
                    lin = [[xa, ya, 0.0], [xb, yb, 0.0]]
                    lines = mirror_8x(lin, origin, line_hor, line_ver, lines)

                if i == j and i < nx - 1:
                    # Main diagonal:
                    xa = x0 + dx*i
                    ya = y0 + dy*j
                    xb = x0 + dx*(i + 1)
                    yb = y0 + dy*(j + 1)
                    lin = [[xa, ya, 0.0], [xb, yb, 0.0]]
                    lines = mirror_4x(lin, line_hor, line_ver, lines)

                # Horizontal Members:
                xa = x0 + dx*i
                ya = y0 + dy*j
                xb = x0 + dx*(i + 1)
                yb = y0 + dy*j
                lin = [[xa, ya, 0.0], [xb, yb, 0.0]]
                lines = mirror_8x(lin, origin, line_hor, line_ver, lines)

                i += 1
                # Vertical Members:
                xa = x0 + dx*i
                ya = y0 + dy*j
                xb = x0 + dx*i
                yb = y0 + dy*(j + 1)

                if i >= nstop and j >= nstop:
                    x_ = xa
                    y_ = y0 + dy * 1 * (i - j) / (nx - j) + dy * j
                    lin = [[xa, ya, 0.0], [x_, y_, 0.0]]
                    lines = mirror_8x(lin, origin, line_hor, line_ver, lines)
                    lin = [[x_, y_, 0.0], [xb, yb, 0.0]]
                    lines = mirror_8x(lin, origin, line_hor, line_ver, lines)
                else:
                    lin = [[xa, ya, 0.0], [xb, yb, 0.0]]
                    lines = mirror_8x(lin, origin, line_hor, line_ver, lines)

                i -= 1

    mesh = Mesh.from_lines(lines, delete_boundary_face=True)
    # compas.datastructures.mesh_delete_duplicate_vertices
    from compas.datastructures import mesh_delete_duplicate_vertices
    mesh_delete_duplicate_vertices(mesh)
    form = cls.from_mesh(mesh)
    # form = cls.from_lines(lines, delete_boundary_face=True)
    gkey_key = form.gkey_key()

    if fix == 'corners':
        form.vertex_attribute(gkey_key[geometric_key([x0, y0, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x0, y1, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x1, y0, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x1, y1, 0.0])], 'is_fixed', True)
    else:
        [bnds] = form.vertices_on_boundaries()
        for key in bnds:
            form.vertex_attribute(key, 'is_fixed', True)
        form.delete_boundary_edges()

    return form


def mirror_4x(line, line_hor, line_ver, lines):
    """ Helper to mirror an object 4 times."""

    lines.append(line)
    a_mirror, b_mirror = mirror_points_line(line, line_hor)
    lines.append([a_mirror, b_mirror])
    a_mirror, b_mirror = mirror_points_line([a_mirror, b_mirror], line_ver)
    lines.append([a_mirror, b_mirror])
    a_mirror, b_mirror = mirror_points_line(line, line_ver)
    lines.append([a_mirror, b_mirror])

    return lines


def mirror_8x(line, origin, line_hor, line_ver, lines):
    """ Helper to mirror an object 8 times."""

    lines = mirror_4x(line, line_hor, line_ver, lines)
    rot = rotate_points_xy(line, math.pi/2, origin=origin)
    lines = mirror_4x(rot, line_hor, line_ver, lines)

    return lines


def append_mirrored_lines(line, list_, line_hor, line_ver):
    """ Helper to mirror an object 8 times and add to the list"""
    mirror_a = mirror_points_line(line, line_hor)
    mirror_b = mirror_points_line(mirror_a, line_ver)
    mirror_c = mirror_points_line(line, line_ver)
    list_.append(mirror_a)
    list_.append(mirror_b)
    list_.append(mirror_c)


def create_cross_with_diagonal(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=10, fix='all'):
    """ Construct a FormDiagram based on cross discretiastion with diagonals.

    Parameters
    ----------
    xy_span : list, optional
        List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.
    """

    if isinstance(discretisation, list):
        discretisation = discretisation[0]

    y1 = float(xy_span[1][1])
    y0 = float(xy_span[1][0])
    x1 = float(xy_span[0][1])
    x0 = float(xy_span[0][0])
    x_span = x1 - x0
    y_span = y1 - y0
    dx = x_span/discretisation
    dy = y_span/discretisation

    lines = []

    for i in range(discretisation+1):
        for j in range(discretisation+1):
            if i < discretisation and j < discretisation:
                # Hor Members:
                xa = x0 + dx*i
                ya = y0 + dy*j
                xb = x0 + dx*(i + 1)
                yb = y0 + dy*j
                # Ver Members:
                xc = x0 + dx*i
                yc = y0 + dy*j
                xd = x0 + dx*i
                yd = y0 + dy*(j + 1)
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
                lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                # lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                if (i < discretisation/2 and j < discretisation/2) or (i >= discretisation/2 and j >= discretisation/2):
                    # Diagonal Members in + Direction:
                    xc = x0 + dx*i
                    yc = y0 + dy*j
                    xd = x0 + dx*(i + 1)
                    yd = y0 + dy*(j + 1)
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                else:
                    # Diagonal Members in - Direction:
                    xc = x0 + dx*i
                    yc = y0 + dy*(j + 1)
                    xd = x0 + dx*(i + 1)
                    yd = y0 + dy*j
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
                    # if i == (discretisation - 1):
                    #     xc = x0 + dx*i
                    #     yc = y0 + dy*j
                    #     xd = x0 + dx*(i + 1)
                    #     yd = y0 + dy*(j - 1)
                    #     lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])
            else:
                if i == discretisation and j < discretisation:
                    # Vertical Members on last column:
                    xa = x0 + dx*j
                    ya = y0 + dy*i
                    xb = x0 + dx*(j + 1)
                    yb = y0 + dy*i
                    # Horizontal Members:
                    xc = x0 + dx*i
                    yc = y0 + dy*j
                    xd = x0 + dx*i
                    yd = y0 + dy*(j + 1)
                    lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])
                    lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])

    mesh = Mesh.from_lines(lines, delete_boundary_face=True)
    form = cls.from_mesh(mesh)
    # form = cls.from_lines(lines, delete_boundary_face=True)
    gkey_key = form.gkey_key()

    if fix == 'corners':
        form.vertex_attribute(gkey_key[geometric_key([x0, y0, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x0, y1, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x1, y0, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x1, y1, 0.0])], 'is_fixed', True)
    else:
        [bnds] = form.vertices_on_boundaries()
        for key in bnds:
            form.vertex_attribute(key, 'is_fixed', True)
        form = form.delete_boundary_edges()  # Check if this should be here, or explicit

    return form


def create_fan_form(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=[10, 10], fix='corners'):
    """ Helper to construct a FormDiagram based on fan discretiastion with straight lines to the corners.

    Parameters
    ----------
    xy_span : list, optional
        List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.
    """

    if isinstance(discretisation, int):
        discretisation = [discretisation, discretisation]
    if discretisation[0] % 2 != 0 or discretisation[1] % 2 != 0:
        msg = "Warning!: discretisation of this form diagram has to be even."
        raise ValueError(msg)

    y1 = float(xy_span[1][1])
    y0 = float(xy_span[1][0])
    x1 = float(xy_span[0][1])
    x0 = float(xy_span[0][0])

    x_span = x1 - x0
    y_span = y1 - y0
    xc0 = x0 + x_span/2
    yc0 = y0 + y_span/2
    division_x = discretisation[0]
    division_y = discretisation[1]
    dx = float(x_span/division_x)
    dy = float(y_span/division_y)
    nx = int(division_x/2)
    ny = int(division_y/2)
    line_hor = [[x0, yc0, 0.0], [xc0, yc0, 0.0]]
    line_ver = [[xc0, y0, 0.0], [xc0, yc0, 0.0]]

    lines = []

    for i in range(nx):
        for j in range(ny+1):
            # Diagonal Members:
            xa = x0 + dx * i
            ya = y0 + dy * j * i / nx
            xb = x0 + dx * (i + 1)
            yb = y0 + dy * j * (i + 1) / nx
            lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            a_mirror, b_mirror = mirror_points_line([[xa, ya, 0.0], [xb, yb, 0.0]], line_hor)
            lines.append([a_mirror, b_mirror])
            a_mirror, b_mirror = mirror_points_line([a_mirror, b_mirror], line_ver)
            lines.append([a_mirror, b_mirror])
            a_mirror, b_mirror = mirror_points_line([[xa, ya, 0.0], [xb, yb, 0.0]], line_ver)
            lines.append([a_mirror, b_mirror])

            xa_ = x0 + dx * j * i / nx
            ya_ = y0 + dy * i
            xb_ = x0 + dx * j * (i + 1) / nx
            yb_ = y0 + dy * (i + 1)
            lines.append([[xa_, ya_, 0.0], [xb_, yb_, 0.0]])

            a_mirror, b_mirror = mirror_points_line([[xa_, ya_, 0.0], [xb_, yb_, 0.0]], line_hor)
            lines.append([a_mirror, b_mirror])
            a_mirror, b_mirror = mirror_points_line([a_mirror, b_mirror], line_ver)
            lines.append([a_mirror, b_mirror])
            a_mirror, b_mirror = mirror_points_line([[xa_, ya_, 0.0], [xb_, yb_, 0.0]], line_ver)
            lines.append([a_mirror, b_mirror])

            if j < ny:
                # Vertical or Horizontal Members:
                xc = x0 + dx * (i + 1)
                yc = y0 + dy * j * (i + 1) / nx
                xd = x0 + dx * (i + 1)
                yd = y0 + dy * (j + 1) * (i + 1) / nx
                lines.append([[xc, yc, 0.0], [xd, yd, 0.0]])

                c_mirror, d_mirror = mirror_points_line([[xc, yc, 0.0], [xd, yd, 0.0]], line_hor)
                lines.append([c_mirror, d_mirror])
                c_mirror, d_mirror = mirror_points_line([c_mirror, d_mirror], line_ver)
                lines.append([c_mirror, d_mirror])
                c_mirror, d_mirror = mirror_points_line([[xc, yc, 0.0], [xd, yd, 0.0]], line_ver)
                lines.append([c_mirror, d_mirror])

                xc_ = x0 + dx * j * (i + 1) / nx
                yc_ = y0 + dy * (i + 1)
                xd_ = x0 + dx * (j + 1) * (i + 1) / nx
                yd_ = y0 + dy * (i + 1)
                lines.append([[xc_, yc_, 0.0], [xd_, yd_, 0.0]])

                c_mirror, d_mirror = mirror_points_line([[xc_, yc_, 0.0], [xd_, yd_, 0.0]], line_hor)
                lines.append([c_mirror, d_mirror])
                c_mirror, d_mirror = mirror_points_line([c_mirror, d_mirror], line_ver)
                lines.append([c_mirror, d_mirror])
                c_mirror, d_mirror = mirror_points_line([[xc_, yc_, 0.0], [xd_, yd_, 0.0]], line_ver)
                lines.append([c_mirror, d_mirror])

    form = cls.from_lines(lines)

    gkey_key = form.gkey_key()

    if fix == 'corners':
        form.vertex_attribute(gkey_key[geometric_key([x0, y0, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x0, y1, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x1, y0, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x1, y1, 0.0])], 'is_fixed', True)
    else:
        [bnds] = form.vertices_on_boundaries()
        for key in bnds:
            form.vertex_attribute(key, 'is_fixed', True)
        form.delete_boundary_edges()

        # Recreate the mesh helps to avoid problems in the future

        vertices, faces = form.to_vertices_and_faces()
        newmesh = Mesh.from_vertices_and_faces(vertices, faces)

        form = cls.from_mesh(newmesh)

        [bnds] = form.vertices_on_boundaries()
        for key in bnds:
            form.vertex_attribute(key, 'is_fixed', True)

        for edge in form.edges_on_boundary():
            form.edge_attribute(edge, '_is_edge', False)

    return form


def create_ortho_form(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=[10, 10], fix='corners'):
    """ Helper to construct a FormDiagram based on a simple orthogonal discretisation.

    Parameters
    ----------
    xy_span : list, optional
        List with initial- and end-points of the vault, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.
    """

    if isinstance(discretisation, int):
        discretisation = [discretisation, discretisation]
    # if discretisation[0] % 2 != 0 or discretisation[1] % 2 != 0:
    #     msg = "Warning!: discretisation of this form diagram has to be even."
    #     raise ValueError(msg)

    y1 = float(xy_span[1][1])
    y0 = float(xy_span[1][0])
    x1 = float(xy_span[0][1])
    x0 = float(xy_span[0][0])
    x_span = x1 - x0
    y_span = y1 - y0
    division_x = discretisation[0]
    division_y = discretisation[1]
    dx = float(x_span/division_x)
    dy = float(y_span/division_y)

    vertices = []
    faces = []

    for j in range(division_y+1):
        for i in range(division_x+1):
            xi = x0 + dx*i
            yi = y0 + dy*j
            vertices.append([xi, yi, 0.0])
            if i < division_x and j < division_y:
                p1 = j * (division_x + 1) + i
                p2 = j * (division_x + 1) + i + 1
                p3 = (j + 1) * (division_x + 1) + i + 1
                p4 = (j + 1) * (division_x + 1) + i
                face = [p1, p2, p3, p4, p1]
                faces.append(face)
                print(face)

    mesh = Mesh.from_vertices_and_faces(vertices, faces)

    form = cls.from_mesh(mesh)
    gkey_key = form.gkey_key()

    if fix == 'corners':
        form.vertex_attribute(gkey_key[geometric_key([x0, y0, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x0, y1, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x1, y0, 0.0])], 'is_fixed', True)
        form.vertex_attribute(gkey_key[geometric_key([x1, y1, 0.0])], 'is_fixed', True)
    else:
        [bnds] = form.vertices_on_boundaries()
        for key in bnds:
            form.vertex_attribute(key, 'is_fixed', True)

        for edge in form.edges_on_boundary():
            form.edge_attribute(edge, '_is_edge', False)

        for vertex in list(form.vertices_where({'vertex_degree': 2})):
            nbrs = form.vertex_neighbors(vertex)
            if all(not form.edge_attribute((vertex, nbr), '_is_edge') for nbr in nbrs):
                for nbr in nbrs:
                    face = form.halfedge[vertex][nbr]
                    if face is not None:
                        break
                vertices = form.face_vertices(face)
                after = nbr
                before = vertices[vertices.index(after) - 2]
                form.split_face(face, before, after)
                form.edge_attribute((before, after), '_is_edge', False)
                form.delete_vertex(vertex)

        vertices, faces = form.to_vertices_and_faces()
        newmesh = Mesh.from_vertices_and_faces(vertices, faces)

        form = cls.from_mesh(newmesh)

        [bnds] = form.vertices_on_boundaries()
        for key in bnds:
            form.vertex_attribute(key, 'is_fixed', True)

        for edge in form.edges_on_boundary():
            form.edge_attribute(edge, '_is_edge', False)

        # form.delete_boundary_edges()

    return form


def create_parametric_form(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=10, lambd=0.5, fix='corners'):
    """Create a parametric form diagram based on the inclination lambda of the arches

    Parameters
    ----------
    xy_span : [[float, float], [float, float]], optional
        List with initial- and end-points of the vault, by default, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    lambd : float, optional
        Inclination of the arches in the diagram (0.0 will result in cross and 1.0 in fan diagrams), by default 0.5
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.

    Notes
    ---------
        Diagram implemented after `N. A. Nodargi et al., 2022 <https://doi.org/10.1016/j.engstruct.2022.114878>`_.
    """

    from compas_tno.utilities import split_intersection_lines
    from compas_tno.utilities import move_pattern_to_origin
    from compas_tno.utilities import mesh_remove_two_valent_nodes

    if 0.0 > lambd or lambd > 1.0:
        raise ValueError('Lambda should be between 0.0 and 1.0')

    x_span = xy_span[0][1] - xy_span[0][0]
    y_span = xy_span[1][1] - xy_span[1][0]
    is_retangular = False

    if abs(x_span - y_span) > 1e-6:
        is_retangular = True
        y_span = x_span = 10.0
        x0, x1 = y0, y1 = 0.0, 10.0
    else:
        x0, x1 = xy_span[0][0], xy_span[0][1]
        y0, y1 = xy_span[1][0], xy_span[1][1]

    xc = (x1 + x0)/2
    yc = (y1 + y0)/2

    xc0 = x0 + x_span/2
    yc0 = y0 + y_span/2
    division_x = discretisation
    division_y = discretisation
    dx = float(x_span/division_x)
    dy = float(y_span/division_y)
    nx = int(division_x/2)
    line_hor = [[x0, yc0, 0.0], [xc0, yc0, 0.0]]
    line_ver = [[xc0, y0, 0.0], [xc0, yc0, 0.0]]

    lines = []

    for i in range(nx + 1):
        j = i

        xa = xc
        ya = xc - dy*j
        xb = (ya - y0) * (1 - lambd)
        yb = (ya - y0) * (1 - lambd)

        if distance_point_point_xy([xa, ya], [xb, yb]):
            lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            append_mirrored_lines([[xa, ya, 0.0], [xb, yb, 0.0]], lines, line_hor, line_ver)

        if i == 0:
            xa = x0
            ya = y0

            if distance_point_point_xy([xa, ya], [xb, yb]):
                lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

                append_mirrored_lines([[xa, ya, 0.0], [xb, yb, 0.0]], lines, line_hor, line_ver)

        xa = xc - dx*j
        xb = xc - dx*j
        ya = yc - dy*j
        yb = y0

        if distance_point_point_xy([xa, ya], [xb, yb]):
            lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            append_mirrored_lines([[xa, ya, 0.0], [xb, yb, 0.0]], lines, line_hor, line_ver)

        xa = yc - dy*j
        ya = yc
        xb = (xa - x0) * (1 - lambd)
        yb = xb * (y1 - y0) / (x1 - x0) + y0

        if distance_point_point_xy([xa, ya], [xb, yb]):
            lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            append_mirrored_lines([[xa, ya, 0.0], [xb, yb, 0.0]], lines, line_hor, line_ver)

        xa = xc - dx*j
        xb = x0
        ya = yc - dy*j
        yb = yc - dy*j

        if distance_point_point_xy([xa, ya], [xb, yb]):
            lines.append([[xa, ya, 0.0], [xb, yb, 0.0]])

            append_mirrored_lines([[xa, ya, 0.0], [xb, yb, 0.0]], lines, line_hor, line_ver)

    clean_lines = split_intersection_lines(lines)

    mesh = Mesh.from_lines(clean_lines, delete_boundary_face=True)
    mesh = mesh_weld(mesh)

    # if valence is 2
    mesh = mesh_remove_two_valent_nodes(mesh, delete_boundary_face=True)

    x0, x1 = xy_span[0][0], xy_span[0][1]
    y0, y1 = xy_span[1][0], xy_span[1][1]
    corners = [[x0, y0], [x1, y0], [x1, y1], [x0, y1]]

    if is_retangular:
        move_pattern_to_origin(mesh, corners=corners)

    vertices, faces = mesh.to_vertices_and_faces()
    form = cls.from_vertices_and_faces(vertices, faces)
    # form = cls.from_mesh(mesh)

    if fix == 'corners':
        for key in form.vertices():
            pt = form.vertex_coordinates(key)
            for corner in corners:
                dist = distance_point_point_xy(pt, corner)
                if dist < 1e-3:
                    form.vertex_attribute(key, 'is_fixed', True)
                    break
    else:
        [bnds] = form.vertices_on_boundaries()
        for key in bnds:
            form.vertex_attribute(key, 'is_fixed', True)
        form = form.delete_boundary_edges()  # Check if this should be here, or explicit

    return form


def create_delta_form(cls, xy_span=[[0.0, 10.0], [0.0, 10.0]], discretisation=10, delta=0.5, fix='corners'):
    """Create a parametric form diagram based on a sag to the interior of the pattern

    Parameters
    ----------
    xy_span : [[float, float], [float, float]], optional
        List with initial- and end-points of the vault, by default, by default [[0.0, 10.0], [0.0, 10.0]]
    discretisation : int, optional
        Set the density of the grid in x and y directions, by default 10
    delta : float, optional
        Sag applied to the boundary of the pattern, by default 0.5
    fix : str, optional
        Option to select the constrained nodes: 'corners', 'all' are accepted, by default 'corners'

    Returns
    -------
    :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram created.

    """

    form = create_cross_with_diagonal(cls, xy_span=xy_span, discretisation=discretisation, fix=fix)

    from compas_tno.utilities.form import slide_pattern_inwards

    slide_pattern_inwards(form, delta=delta)

    return form



import math

from compas.data import Data

from compas_tno.problems import set_up_convex_optimisation
from compas_tno.problems import set_up_general_optimisation

from compas_tno.solvers import run_optimisation_scipy
from compas_tno.solvers import run_optimisation_MATLAB
from compas_tno.solvers import run_optimisation_CVXPY
from compas_tno.solvers import run_optimisation_MMA
from compas_tno.solvers import run_optimisation_ipopt

from compas_tno.utilities import apply_selfweight_from_shape
from compas_tno.utilities import apply_selfweight_from_pattern
from compas_tno.utilities import get_shape_ub
from compas_tno.utilities import get_shape_lb

from compas_tno.utilities import apply_horizontal_multiplier
from compas_tno.utilities import apply_envelope_from_shape
from compas_tno.utilities import apply_bounds_on_q
from compas_tno.utilities import apply_bounds_reactions
from compas_tno.utilities import apply_envelope_on_xy

from compas_tno.diagrams import FormDiagram
from compas_tno.shapes import Shape
from compas_tno.optimisers import Optimiser

from numpy import array
from scipy import interpolate


class Analysis(Data):
    """The ``Analysis`` class unites the :class:`~compas_tno.diagrams.FormDiagram`, the :class:`~compas_tno.shapes.Shape`
    and the :class:`~compas_tno.optimisers.Optimiser` to compute the optimisation.

    Examples
    --------
    >>> from compas_tno.analysis import Analysis
    >>> from compas_tno.shapes import Shape
    >>> from compas_tno.diagrams import FormDiagram
    >>> arch = Shape.create_arch()
    >>> form = FormDiagram.create_arch()
    >>> analysis = Analysis.create_minthrust_analysis(form, arch)

    """

    def __init__(self, name="Analysis"):
        self.settings = {}
        self.form = None
        self.shape = None
        self.optimiser = None
        self.name = name

    @property
    def data(self):
        """dict : A data dict representing the shape data structure for serialization.
        """
        data = {
            'form': self.form.data,
            'optimiser': self.optimiser.data,
            'shape': self.shape.data,
            'name': self.name,
            'settings': self.settings
        }
        return data

    @data.setter
    def data(self, data):
        if 'data' in data:
            data = data['data']
        self.settings = data.get('settings') or {}
        self.name = data.get('name')

        formdata = data.get('form', None)
        shapedata = data.get('shape', None)
        optimiserdata = data.get('optimiser', None)

        self.form = None
        self.shape = None
        self.optimiser = None

        if formdata:
            self.form = FormDiagram.from_data(formdata)
        if shapedata:
            self.shape = Shape.from_data(shapedata)
        if optimiserdata:
            self.optimiser = Optimiser.from_data(optimiserdata)

    @classmethod
    def from_elements(cls, shape, form, optimiser):
        """Create a analysis from the elements of the problem (form, shape and optimiser)

        Parameters
        ----------
        shape : :class:`~compas_tno.shapes.Shape`
            The vaulted structure constraining the problem
        form : :class:`~compas_tno.diagrams.FormDiagram`
            The form diagram to analyse the structure
        optimiser : :class:`~compas_tno.optimisers.Optimiser`
            The optimiser with information about the problem

        Returns
        -------
        analysis: Analysis
            The analysis object

        """

        analysis = cls()
        analysis.shape = shape
        analysis.form = form
        analysis.optimiser = optimiser

        return analysis

    @classmethod
    def from_form_and_optimiser(cls, form, optimiser):
        """Create a analysis from the elements of the problem (form and optimiser)"""

        analysis = cls()
        analysis.shape = None
        analysis.form = form
        analysis.optimiser = optimiser

        return analysis

    @classmethod
    def from_form_and_shape(cls, form, shape):
        """Create a analysis from the elements of the problem (form and shape) """

        analysis = cls()
        analysis.shape = shape
        analysis.form = form
        analysis.optimiser = None

        return analysis

    @classmethod
    def create_minthk_analysis(cls, form, shape, printout=False, plot=False, max_iter=500, starting_point='loadpath', solver='SLSQP', derivatives=True):
        """Create a minimum thickness analysis from the elements of the problem (form and shape)

        Parameters
        ----------
        form : :class:`~compas_tno.diagrams.FormDiagram`
            _description_
        shape : :class:`~compas_tno.shapes.Shape`
            The shape cconstraining the problem
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'

        Returns
        -------
        analysis: Analysis
            The Analysis object

        """

        analysis = cls().from_form_and_shape(form, shape)

        optimiser = Optimiser.create_minthk_optimiser(printout=printout,
                                                      plot=plot,
                                                      max_iter=max_iter,
                                                      starting_point=starting_point,
                                                      solver=solver,
                                                      derivatives=derivatives)

        if printout:
            print('-'*20)
            print('Minimum thickness analysis created')
            print(optimiser)

        analysis.optimiser = optimiser

        return analysis

    @classmethod
    def create_bestfit_analysis(cls, form, shape, printout=False, plot=False, max_iter=500, starting_point='loadpath', solver='SLSQP', derivatives=True):
        """Create a bestfit analysis from the elements of the problem (form and shape)

        Parameters
        ----------
        form : :class:`~compas_tno.diagrams.FormDiagram`
            _description_
        shape : :class:`~compas_tno.shapes.Shape`
            The shape cconstraining the problem
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'
        derivatives : bool, optional
            Whether or not derivatives should be considered, by default False

        Returns
        -------
        analysis: Analysis
            The Analysis object

        """

        analysis = cls().from_form_and_shape(form, shape)

        optimiser = Optimiser.create_bestfit_optimiser(printout=printout,
                                                       plot=plot,
                                                       max_iter=max_iter,
                                                       starting_point=starting_point,
                                                       solver=solver,
                                                       derivatives=derivatives)

        if printout:
            print('-'*20)
            print('Minimum thickness analysis created')
            print(optimiser)

        analysis.optimiser = optimiser

        return analysis

    @classmethod
    def create_minthrust_analysis(cls, form, shape, printout=False, plot=False, max_iter=500, starting_point='loadpath', solver='SLSQP'):
        """Create a minimum thickness analysis from the elements of the problem (form and shape)

        Parameters
        ----------
        form : :class:`~compas_tno.diagrams.FormDiagram`
            _description_
        shape : :class:`~compas_tno.shapes.Shape`
            The shape cconstraining the problemf
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'

        Returns
        -------
        analysis: Analysis
            The Analysis object

        """

        analysis = cls().from_form_and_shape(form, shape)

        optimiser = Optimiser.create_minthrust_optimiser(printout=printout,
                                                         plot=plot,
                                                         max_iter=max_iter,
                                                         starting_point=starting_point,
                                                         solver=solver)

        if printout:
            print('-'*20)
            print('Minimum thrust analysis created')
            print(optimiser)

        analysis.optimiser = optimiser

        return analysis

    @classmethod
    def create_maxthrust_analysis(cls, form, shape, printout=False, plot=False, max_iter=500, starting_point='loadpath', solver='SLSQP'):
        """Create a maximum thickness analysis from the elements of the problem (form and shape)

        Parameters
        ----------
        form : :class:`~compas_tno.diagrams.FormDiagram`
            _description_
        shape : :class:`~compas_tno.shapes.Shape`
            The shape cconstraining the problemf
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'

        Returns
        -------
        analysis: Analysis
            The Analysis object

        """

        analysis = cls().from_form_and_shape(form, shape)

        optimiser = Optimiser.create_maxhrust_optimiser(printout=printout,
                                                        plot=plot,
                                                        max_iter=max_iter,
                                                        starting_point=starting_point,
                                                        solver=solver)

        if printout:
            print('-'*20)
            print('Maximium thrust analysis created')
            print(optimiser)

        analysis.optimiser = optimiser

        return analysis

    @classmethod
    def create_max_load_analysis(cls, form, shape, printout=False, plot=False, horizontal=False, max_iter=500, starting_point='loadpath',
                                 solver='IPOPT', derivatives=True, load_direction=None, max_lambd=1.0):
        """Create a minimum thickness analysis from the elements of the problem (form and shape)

        Parameters
        ----------
        form : :class:`~compas_tno.diagrams.FormDiagram`
            _description_
        shape : :class:`~compas_tno.shapes.Shape`
            The shape cconstraining the problem
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        plot : bool, optional
            Whether or load applied is horizontal, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'

        Returns
        -------
        analysis: Analysis
            The Analysis object

        """

        analysis = cls().from_form_and_shape(form, shape)

        if horizontal:
            optimiser = Optimiser.create_max_horload_optimiser(printout=printout,
                                                               plot=plot,
                                                               max_iter=max_iter,
                                                               starting_point=starting_point,
                                                               solver=solver,
                                                               derivatives=derivatives,
                                                               load_direction=load_direction,
                                                               max_lambd=max_lambd)

            if printout:
                print('-'*20)
                print('Max horizontal load analysis created')
                # print(optimiser)

        else:
            optimiser = Optimiser.create_max_vertload_optimiser(printout=printout,
                                                                plot=plot,
                                                                max_iter=max_iter,
                                                                starting_point=starting_point,
                                                                solver=solver,
                                                                derivatives=derivatives,
                                                                load_direction=load_direction,
                                                                max_lambd=max_lambd)

            if printout:
                print('-'*20)
                print('Max vertical load analysis created')
                # print(optimiser)

        analysis.optimiser = optimiser

        return analysis

    @classmethod
    def create_compl_energy_analysis(cls, form, shape, printout=False, solver='IPOPT', plot=False, max_iter=500, starting_point='loadpath',
                                     support_displacement=None, Emethod='simplified'):
        """Create a complementary energy analysis from the elements of the problem (form and shape)

        Parameters
        ----------
        form : :class:`~compas_tno.diagrams.FormDiagram`
            _description_
        shape : :class:`~compas_tno.shapes.Shape`
            The shape cconstraining the problemf
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'

        Returns
        -------
        analysis: Analysis
            The Analysis object

        """

        analysis = cls().from_form_and_shape(form, shape)

        optimiser = Optimiser.create_compl_energy_optimiser(printout=printout,
                                                            plot=plot,
                                                            max_iter=max_iter,
                                                            starting_point=starting_point,
                                                            support_displacement=support_displacement,
                                                            Emethod=Emethod,
                                                            solver=solver)

        if printout:
            print('-'*20)
            print('Complementary energy created')
            print(optimiser)

        analysis.optimiser = optimiser

        return analysis

    @classmethod
    def create_quad_compl_energy_analysis(cls, form, shape, printout=False, solver='IPOPT', plot=False, max_iter=500,
                                          starting_point='loadpath', support_displacement=None, Emethod='simplified'):
        """Create a complementary energy analysis including a quadratic term from the elements of the problem (form and shape)

        Parameters
        ----------
        form : :class:`~compas_tno.diagrams.FormDiagram`
            _description_
        shape : :class:`~compas_tno.shapes.Shape`
            The shape cconstraining the problemf
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'

        Returns
        -------
        analysis: Analysis
            The Analysis object

        """

        analysis = cls().from_form_and_shape(form, shape)

        optimiser = Optimiser.create_compl_energy_optimiser(printout=printout,
                                                            plot=plot,
                                                            max_iter=max_iter,
                                                            starting_point=starting_point,
                                                            support_displacement=support_displacement,
                                                            Emethod=Emethod,
                                                            solver=solver)

        if printout:
            print('-'*20)
            print('Complementary energy analysis created')
            print(optimiser)

        analysis.optimiser = optimiser

        return analysis

    @classmethod
    def create_lp_analysis(cls, form, shape=None, solver='CVXPY', printout=False, plot=False, max_iter=500):
        """Create a minimum thickness analysis from the elements of the problem (form and shape)

        Parameters
        ----------
        form : :class:`~compas_tno.diagrams.FormDiagram`
            _description_
        shape : :class:`~compas_tno.shapes.Shape`, optional
            The shape cconstraining the problem, by default None
        printout : bool, optional
            Whether or not prints appear in the creen, by default False
        plot : bool, optional
            Whether or not plots showing intermediate states appear, by default False
        max_iter : int, optional
            Maximum number of itetations, by default 500
        starting_point : str, optional
            Which starting point use, by default 'loadpath'

        Returns
        -------
        analysis: Analysis
            The Analysis object

        """

        optimiser = Optimiser.create_lp_optimiser(solver=solver,
                                                  printout=printout,
                                                  plot=plot,
                                                  max_iter=max_iter)

        if shape:
            analysis = cls().from_elements(shape, form, optimiser)
        else:
            analysis = cls().from_form_and_optimiser(form, optimiser)

        if printout:
            print('-'*20)
            print('Load path Analysis created')
            print(optimiser)

        analysis.optimiser = optimiser

        return analysis

    def is_convex(self):
        """Check if the analysis problem is convex."""

        if not self.optimiser:
            raise ValueError('Define the Optimiser for the problem')

        objective = self.optimiser.settings['objective']
        features = self.optimiser.settings['features']

        if objective == 'loadpath' and 'fixed' in features:
            return True
        else:
            return False

    def set_optimiser_options(self, **kwargs):
        """Set the additional options of the optimisation.
        """

        if kwargs:
            self.optimiser.set_additional_options(**kwargs)

    def clear_previous_results(self):
        """Clear Previous results stored in the Analysis object. Necessary to perform sequential optimisation with the same analysis object
        """

        self.optimiser.clear_optimiser()

        return

    def apply_selfweight(self, normalize_loads=True):
        """Invoke method to apply selfweight to the nodes of the form diagram based on the shape"""
        apply_selfweight_from_shape(self.form, self.shape, normalize=normalize_loads)

        return

    def apply_selfweight_from_pattern(self, pattern, plot=False):
        """Apply selfweight to the nodes considering a different Form Diagram to locate loads.

            Warning, the base pattern has to coincide with nodes from the original form diagram.
        """

        apply_selfweight_from_pattern(self.form, pattern, plot=plot)

        return

    def apply_hor_multiplier(self, multiplier, component):
        """Apply a multiplier on the selfweight to the nodes of the form diagram based"""

        apply_horizontal_multiplier(self.form, lambd=multiplier, direction=component)

        return

    def apply_envelope(self):
        """Invoke method to apply ub and lb to the nodes based on the shape's intrados and extrados"""

        apply_envelope_from_shape(self.form, self.shape)

        return

    def apply_bounds_on_q(self, qmax=0.0, qmin=-10000.0):
        """Invoke method to apply bounds on the force densities of the pattern (qmax, qmin)"""

        apply_bounds_on_q(self.form, qmax=qmax, qmin=qmin)

        return

    def apply_envelope_with_damage(self):
        """Apply ub and lb to the nodes based on the shape's intrados and extrados and in the intra/extra damaged"""

        form = self.form
        shape = self.shape
        extrados_damage = array(shape.extrados_damage.vertices_attributes('xyz'))
        intrados_damage = array(shape.intrados_damage.vertices_attributes('xyz'))

        for key in form.vertices():
            x, y, _ = form.vertex_coordinates(key)
            ub_ = get_shape_ub(shape, x, y)
            lb_ = get_shape_lb(shape, x, y)
            lb_damage = float(interpolate.griddata(intrados_damage[:, :2], intrados_damage[:, 2], [x, y]))
            ub_damage = float(interpolate.griddata(extrados_damage[:, :2], extrados_damage[:, 2], [x, y]))
            if math.isnan(lb_damage) or math.isnan(lb_):
                lb_ = -1 * shape.datashape['t']
                # print('Shape interpolation got NaN, check results (x,y): ({0:.3f}, {1:.3f})'.format(x, y))
            else:
                lb_ = max(lb_, lb_damage)
            ub_ = min(ub_, ub_damage)
            form.vertex_attribute(key, 'ub', value=ub_)
            form.vertex_attribute(key, 'lb', value=lb_)

        self.form = form  # With correct forces

        return

    def apply_target(self):
        """Apply target to the nodes based on the shape's target surface"""

        form = self.form
        shape = self.shape

        for key in form.vertices():
            x, y, _ = form.vertex_coordinates(key)
            form.vertex_attribute(key, 'target', value=shape.get_middle(x, y))

        # Go over nodes and find node = key and apply the pointed load pz += magnitude

        self.form = form  # With correct forces

        return

    def apply_envelope_on_xy(self, c=0.5):
        """_summary_

        Parameters
        ----------
        c : float, optional
            Distance in (x, y) to constraint the nodes limiting the hor. movement, by default 0.5

        """

        apply_envelope_on_xy(self.form, c=c)

        return

    def apply_cracks(self, key, position):
        """Apply cracks on the nodes (key) and in the positions up / down"""

        form = self.form

        # Go over nodes and find node = key and apply the pointed load pz += magnitude

        self.form = form  # With correct forces

        return

    def apply_reaction_bounds(self, assume_shape=None):
        """Apply limit thk to be respected by the anchor points"""

        apply_bounds_reactions(self.form, self.shape, assume_shape)

        return

    def set_up_optimiser(self):
        """With the data from the elements of the problem compute the matrices for the optimisation"""

        if self.is_convex():
            set_up_convex_optimisation(self)
        else:
            self = set_up_general_optimisation(self)

        return

    def run(self):
        """With the data from the elements of the problem compute the matrices for the optimisation"""

        solver = self.optimiser.settings.get('solver', 'SLSQP')

        if not isinstance(solver, str):
            raise ValueError('Please provide the name of the solver')

        if self.is_convex():
            if solver == 'MATLAB':
                run_optimisation_MATLAB(self)
            elif solver == 'CVXPY':
                run_optimisation_CVXPY(self)
            else:
                raise NotImplementedError('Only <CVXPY> and <MATLAB> are suitable for this optimisation')
        else:
            if solver.split('-') == 'pyOpt':
                self = run_optimisation_MMA(self)  # change to PyOpt
            elif solver == 'MMA':
                self = run_optimisation_MMA(self)
            elif solver == 'IPOPT':
                self = run_optimisation_ipopt(self)
            else:
                self = run_optimisation_scipy(self)

        return

    def __str__(self):
        tpl = "<Analysis with parameters: {} >".format(self.data)
        return tpl



from compas_tno.shapes import Shape
from compas_tno.diagrams import FormDiagram
from compas_tno.viewers import Viewer
from compas_tno.analysis import Analysis


# ----------------------------------------
# 1. Shape geometric definition
# ----------------------------------------
radius = 5.0
thk = 0.50
center = [5, 5]
dome = Shape.create_dome(radius=radius, thk=thk, center=center)
dome.ro = 1.0

# ----------------------------------------
# 2. Form diagram geometric definition
# ----------------------------------------
discretisation = [16, 20]
form = FormDiagram.create_circular_radial_form(radius=radius, discretisation=discretisation)

# --------------------------------------------
# 3. Minimum thurst solution and visualisation
# --------------------------------------------
analysis = Analysis.create_minthk_analysis(form, dome, solver='IPOPT', printout=True)
analysis.optimiser.set_constraints(['funicular', 'envelope', 'reac_bounds'])
analysis.apply_selfweight()
analysis.apply_envelope()
analysis.apply_reaction_bounds()
analysis.set_up_optimiser()
analysis.run()

thk_min = analysis.optimiser.fopt
dome_min = Shape.create_dome(radius=radius, thk=thk_min, center=center)

view = Viewer(form, dome_min)
view.scale_edge_thickness(5.0)
view.draw_form()
view.draw_shape()
view.draw_reactions(extend_reactions=True)
view.draw_cracks()
view.show()



from compas_tno.shapes import MeshDos
from compas.datastructures import mesh_weld
from numpy import ones
from numpy import zeros
from numpy import array
import math


def arch_shape(H=1.00, L=2.0, x0=0.0, thk=0.20, b=0.5, t=0.0, total_nodes=100):
    """Helper to create meshes to define upper and lower bounds of 2D arch

    Parameters
    ----------
    H : float, optional
        Height of the arch, by default 1.00
    L : float, optional
        Span of the arch, by default 2.0
    x0 : float, optional
        Starting coordinate of the arch , by default 0.0
    thk : float, optional
        Thickness of the arch, by default 0.20
    b : float, optional
        Out-of-plane measure of the arch, by default 0.5
    t : float, optional
        Parameter for lower bound in nodes in the boundary, by default 0.0
    total_nodes : int, optional
        Density of the shape, by default 100

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the pattern
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the pattern
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the pattern

    """
    # Add option for starting from Hi and Li for a given thk.

    radius = H / 2 + (L**2 / (8 * H))
    ri = radius - thk/2
    re = radius + thk/2

    spr = math.atan2((L/2), (radius - H))
    tot_angle = 2*spr
    angle_init = (math.pi - tot_angle)/2
    an = tot_angle / (total_nodes - 1)
    zc = radius - H
    xc = L/2 + x0
    i = 0

    xs = []
    ys = []
    zts = []
    zis = []
    zes = []
    faces = []

    for i in range(total_nodes-1):
        angle_i = angle_init + i * an
        angle_f = angle_init + (i + 1) * an
        xi = xc - radius * math.cos(angle_i)
        xf = xc - radius * math.cos(angle_f)
        zi = radius * math.sin(angle_i) - zc
        zf = radius * math.sin(angle_f) - zc
        zei = math.sqrt(re**2 - (xi - xc)**2) - zc
        zef = math.sqrt(re**2 - (xf - xc)**2) - zc
        zii2 = ri**2 - (xi - xc)**2
        zif2 = ri**2 - (xf - xc)**2
        if zii2 > 0:
            zii = math.sqrt(ri**2 - (xi - xc)**2) - zc
        else:
            zii = -t
        if zif2 > 0:
            zif = math.sqrt(ri**2 - (xf - xc)**2) - zc
        else:
            zif = -t

        xs += [xi, xi, xi, xf, xf, xf]
        ys += [-b/2, 0, b/2, -b/2, 0, b/2]
        zts += [zi, zi, zi, zf, zf, zf]
        zis += [zii, zii, zii, zif, zif, zif]
        zes += [zei, zei, zei, zef, zef, zef]
        faces.append([6*i, 6*i + 1, 6*i + 4, 6*i + 3])
        faces.append([6*i + 1, 6*i + 2, 6*i + 5, 6*i + 4])

        i = i + 1

    intrados = mesh_weld(MeshDos.from_vertices_and_faces(array([xs, ys, zis]).transpose(), faces))
    extrados = mesh_weld(MeshDos.from_vertices_and_faces(array([xs, ys, zes]).transpose(), faces))
    middle = mesh_weld(MeshDos.from_vertices_and_faces(array([xs, ys, zts]).transpose(), faces))

    return intrados, extrados, middle


def arch_shape_polar(H=1.00, L=2.0, x0=0.0, thk=0.20, b=0.5, total_nodes=100):
    """Helper to create meshes to define upper and lower bounds of 2D arch with polar coordinates
    Note: The points in this mesh will not result in a direct projection for a typical form diagram

    Parameters
    ----------
    H : float, optional
        Height of the arch, by default 1.00
    L : float, optional
        Span of the arch, by default 2.0
    x0 : float, optional
        Starting coordinate of the arch , by default 0.0
    thk : float, optional
        Thickness of the arch, by default 0.20
    b : float, optional
        Out-of-plane measure of the arch, by default 0.5
    total_nodes : int, optional
        Density of the shape, by default 100

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the pattern
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the pattern
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the pattern

    """
    # Add option for starting from Hi and Li for a given thk.

    radius = H / 2 + (L**2 / (8 * H))
    ri = radius - thk/2
    re = radius + thk/2

    spr = math.atan2((L/2), (radius - H))
    tot_angle = 2*spr
    angle_init = (math.pi - tot_angle)/2
    an = tot_angle / (total_nodes - 1)
    zc = radius - H
    xc = L/2 + x0
    i = 0

    xis = []
    xes = []
    xts = []
    ys = []
    zts = []
    zis = []
    zes = []
    faces = []

    for i in range(total_nodes-1):
        angle_i = angle_init + i * an
        angle_f = angle_init + (i + 1) * an

        xii = xc - ri * math.cos(angle_i)  # intrados initial
        xif = xc - ri * math.cos(angle_f)  # intrados final
        xei = xc - re * math.cos(angle_i)  # extrados initial
        xef = xc - re * math.cos(angle_f)  # extrados final
        xti = xc - radius * math.cos(angle_i)  # target initial
        xtf = xc - radius * math.cos(angle_f)  # target final

        zii = zc + ri * math.sin(angle_i)  # intrados initial
        zif = zc + ri * math.sin(angle_f)  # intrados final
        zei = zc + re * math.sin(angle_i)  # extrados initial
        zef = zc + re * math.sin(angle_f)  # extrados final
        zti = zc + radius * math.sin(angle_i)  # target initial
        ztf = zc + radius * math.sin(angle_f)  # target final

        xis += [xii, xii, xii, xif, xif, xif]
        xes += [xei, xei, xei, xef, xef, xef]
        xts += [xti, xti, xti, xtf, xtf, xtf]
        ys += [-b/2, 0, b/2, -b/2, 0, b/2]
        zts += [zti, zti, zti, ztf, ztf, ztf]
        zis += [zii, zii, zii, zif, zif, zif]
        zes += [zei, zei, zei, zef, zef, zef]
        faces.append([6*i, 6*i + 1, 6*i + 4, 6*i + 3])
        faces.append([6*i + 1, 6*i + 2, 6*i + 5, 6*i + 4])

        i = i + 1

    intrados = mesh_weld(MeshDos.from_vertices_and_faces(array([xis, ys, zis]).transpose(), faces))
    extrados = mesh_weld(MeshDos.from_vertices_and_faces(array([xes, ys, zes]).transpose(), faces))
    middle = mesh_weld(MeshDos.from_vertices_and_faces(array([xts, ys, zts]).transpose(), faces))

    return intrados, extrados, middle


def arch_ub_lb_update(x, y, thk, t, H=1.00, L=2.0, x0=0.0):
    """Update upper and lower bounds of an arch based in the parameters

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    H : float, optional
        Height of the arch, by default 1.00
    L : float, optional
        Span of the arch, by default 2.0
    x0 : float, optional
        Starting coordinate of the arch , by default 0.0

    Returns
    -------
    ub : array
        Values of the upper bound in the points
    lb : array
        Values of the lower bound in the points
    """

    radius = H / 2 + (L**2 / (8 * H))
    ri = radius - thk/2
    re = radius + thk/2
    ub = ones((len(x), 1))
    lb = ones((len(x), 1)) * - t
    zc = radius - H
    xc = L/2 + x0

    for i in range(len(x)):
        ub[i] = math.sqrt(re**2 - (x[i] - xc)**2) - zc
        lb2 = ri**2 - (x[i] - xc)**2
        if lb2 > 0:
            lb[i] = math.sqrt(lb2) - zc

    return ub, lb


def arch_dub_dlb(x, y, thk, t, H=1.00, L=2.0, x0=0.0):
    """Computes the sensitivities of upper and lower bounds in the x, y coordinates and thickness specified.

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    H : float, optional
        Height of the arch, by default 1.00
    L : float, optional
        Span of the arch, by default 2.0
    x0 : float, optional
        Starting coordinate of the arch , by default 0.0

    Returns
    -------
    dub : array
        Values of the sensitivities for the upper bound in the points
    dlb : array
        Values of the sensitivities for the lower bound in the points
    """

    radius = H / 2 + (L**2 / (8 * H))
    ri = radius - thk/2
    re = radius + thk/2
    ub = ones((len(x), 1))
    lb = ones((len(x), 1)) * - t
    dub = zeros((len(x), 1))
    dlb = zeros((len(x), 1))
    zc = radius - H
    xc = L/2 + x0

    for i in range(len(x)):
        ze = math.sqrt(re**2 - (x[i] - xc)**2) - zc
        ub[i] = ze
        dub[i] = re/(2*ze)
        zi2 = ri**2 - (x[i] - xc)**2
        if zi2 > 0:
            zi = math.sqrt(zi2) - zc
            lb[i] = zi
            dlb[i] = - ri/(2*zi)

    return dub, dlb  # ub, lb


def arch_b_update(x, y, thk, fixed, H=1.00, L=2.0, x0=0.0):
    """Computes the sensitivities of upper and lower bounds in the x, y coordinates and thickness specified.

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    H : float, optional
        Height of the arch, by default 1.00
    L : float, optional
        Span of the arch, by default 2.0
    x0 : float, optional
        Starting coordinate of the arch , by default 0.0

    Returns
    -------
    b : array
        Value of the ``b`` parameter.
    """

    radius = H / 2 + (L**2 / (8 * H))
    re = radius + thk/2
    zc = radius - H
    b = zeros((len(fixed), 2))
    x_lim = math.sqrt(re**2 - zc**2)

    for i in range(len(fixed)):
        b[i, :] = [x_lim - L/2, 0.0]

    return b


def arch_db(x, y, thk, fixed, H=1.00, L=2.0, x0=0.0):
    """Computes the sensitivities of upper and lower bounds in the x, y coordinates and thickness specified.

    Parameters
    ----------
    x : list
        x-coordinates of the points
    y : list
        y-coordinates of the points
    thk : float
        Thickness of the arch
    t : float
        Parameter for lower bound in nodes in the boundary
    H : float, optional
        Height of the arch, by default 1.00
    L : float, optional
        Span of the arch, by default 2.0
    x0 : float, optional
        Starting coordinate of the arch , by default 0.0

    Returns
    -------
    db : array
        Sensitivity of the ``b`` parameter.
    """

    radius = H / 2 + (L**2 / (8 * H))
    re = radius + thk/2
    zc = radius - H
    db = zeros((len(fixed), 2))
    x_lim = math.sqrt(re**2 - zc**2)

    for i in range(len(fixed)):
        db[i, :] = [1/2 * re/x_lim, 0.0]

    return db



import compas
from compas.datastructures import Network
from compas.datastructures import network_is_crossed


class FormGraph(Network):
    """A graph representing the geometry and connectivity of the lines of a form diagram.

    Parameters
    ----------


    Attributes
    ----------


    """

    def __init__(self):
        super(FormGraph, self).__init__()

    def node_index(self):
        return {node: index for index, node in enumerate(self.nodes())}

    def is_2d(self):
        """Verify that all nodes of the graph lie in a horizontal plane.

        Returns
        -------
        bool
        """
        z = self.nodes_attribute('z')
        zmin = min(z)
        zmax = max(z)
        if (zmax - zmin) ** 2 > 0.001:
            return False
        return True

    # def make_2d(self):
    #     self.nodes_attribute('z', 0.0)

    def is_planar(self):
        """Verify that the graph has a planar embedding.

        Returns
        -------
        bool
        """
        if compas.IPY:
            from compas.rpc import Proxy
            planarity = Proxy('planarity')
        else:
            import planarity
        return planarity.is_planar(list(self.edges()))

    def is_crossed(self):
        """Verify that the current embedding of the graph has crossing edges.

        Returns
        -------
        bool
        """
        return network_is_crossed(self)

    def is_planar_embedding(self):
        """Verify that the current embedding of the graph is planar."""
        return self.is_2d() and self.is_planar() and not self.is_crossed()

    def embed(self, fixed=None, straightline=True):
        """Compute a geometry for the graph that embeds it in the plane.

        Returns
        -------
        bool
        """
        if compas.IPY:
            from compas.rpc import Proxy
            proxy = Proxy('compas.datastructures')

            def network_embed_in_plane(network, fixed, straightline):
                network.data = proxy.network_embed_in_plane_proxy(network.data, fixed, straightline)

        else:
            from compas.datastructures import network_embed_in_plane
        return network_embed_in_plane(self, fixed, straightline)


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino

from compas_tno.shapes import Shape


__commandname__ = "TNO_shape_from_library"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']

    answer = compas_rhino.rs.GetString("Base geometry of the Shape", "Cancel", ["Circular", "Rectangular", "Cancel"])
    if not answer:
        return
    if answer == "Cancel":
        return

    data = {}
    if answer == "Circular":
        center = compas_rhino.rs.GetPoint(message='Select the center of the diagram')
        if not center:
            return
        data['center'] = [center[0], center[1]]

        radius = compas_rhino.rs.GetReal("Set Radius", 5.0)
        if not radius:
            radius = compas_rhino.rs.GetDistance(first_pt=center, second_pt_msg='Select Second Point to find radius')
        if not radius:
            return
        data['radius'] = radius

        layout = compas_rhino.rs.GetString("Set the type of shape (See docs)", "Cancel", ["dome", "dome_polar", "dome_spr", "Cancel"])
        if not layout:
            return
        if layout == "Cancel":
            return
        data['type'] = layout

        thk = compas_rhino.rs.GetReal("Select the thickness", 0.5)
        if not thk:
            return
        data['thk'] = thk

        discr_1 = compas_rhino.rs.GetInteger(message="Radial discretisaton (#hoops)", number=8)
        if not discr_1:
            return
        discr_2 = compas_rhino.rs.GetInteger(message="Meridian discretisaton", number=12)
        if not discr_2:
            return
        data['discretisation'] = [discr_1, discr_2]

        oculus = compas_rhino.rs.GetString("Add Oculus?", "False", ["True", "False", "Cancel"])
        if oculus == "True":
            radius_oc = compas_rhino.rs.GetReal("Radius of oculus (<{0})".format(radius), radius/4)
            if radius_oc:
                data['r_oculus'] = radius_oc

        t = compas_rhino.rs.GetReal("Define t (See docs.)", 0.0)
        if t:
            data['t'] = t

    if answer == "Rectangular":
        rect = compas_rhino.rs.GetRectangle(mode=1, prompt1='Select the first corner of the diagram', prompt2='Select the second corner of the diagram')
        if not rect:
            return
        xy_coords = [[], []]
        for corner in rect:
            xy_coords[0].append(corner[0])
            xy_coords[1].append(corner[1])

        data['xy_span'] = [[min(xy_coords[0]), max(xy_coords[0])], [min(xy_coords[1]), max(xy_coords[1])]]

        layout = compas_rhino.rs.GetString("Set the type of shape (See docs)", "Cancel", [
                                           "crossvault", "pavillionvault", "parabolic_shell", "pointed_crossvault", "domicalvault", "Cancel"])
        if not layout:
            return
        if layout == "Cancel":
            return
        data['type'] = layout

        thk = compas_rhino.rs.GetReal("Select the thickness", 0.5)
        if not thk:
            return
        data['thk'] = thk

        discr = compas_rhino.rs.GetInteger(message="Select the discretisation", number=10)
        if not discr:
            return
        data['discretisation'] = discr

        t = compas_rhino.rs.GetReal("Select parameter t", 0.0)
        data['t'] = t

        if data['type'] == 'pointed_crossvault':
            min_hc = max((data['xy_span'][0][1] - data['xy_span'][0][0])/2, (data['xy_span'][1][1] - data['xy_span'][1][0])/2)
            hc = compas_rhino.rs.GetReal("Height at midspan (>{0})".format(min_hc), min_hc)
            if not hc:
                return
            data['hc'] = float(max(hc, min_hc))

            add_h = compas_rhino.rs.GetString("Set heights on edge and rubble?", "No", ["Yes", "No", "Cancel"])
            if add_h == "Yes":
                he = compas_rhino.rs.GetReal("Height at edge (<{0})".format(min_hc), min_hc)
                if he:
                    data['he'] = he
                hm = compas_rhino.rs.GetReal("Height at rubble (<{0})".format(min_hc), min_hc)
                if hm:
                    data['hm'] = hm
            else:
                data['hm'] = None
                data['he'] = None

    scene.clear()

    shape = Shape.from_library_proxy(data)

    objects = scene.find_by_name('Shape')
    if not objects:
        scene.add(shape, name='Shape', layer='TNO::Shape')
    else:
        shapeobject = objects[0]
        shapeobject.shape = shape

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas_rhino
from compas_ui.objects import Object
from compas.geometry import Point


__all__ = ['ShapeObject']


class ShapeObject(Object):
    """A form object represents a Shape in the Rhino model space.
    """

    SETTINGS = {
        'show.intrados': True,
        'show.extrados': True,
        'show.middle': False,

        'color.intrados': (0, 0, 0),
        'color.extrados': (0, 0, 0),
        'color.middle': (0, 0, 0),
    }

    def __init__(self, shape, scene=None, name=None, layer=None, visible=True, settings=None):
        super(ShapeObject, self).__init__(shape, scene, name, layer, visible)
        self.settings.update(ShapeObject.SETTINGS)
        self._shape = shape
        self._guids = []
        self._guid_intrados = {}
        self._guid_extrados = {}
        self._guid_middle = {}
        self._location = None
        self._scale = None
        self._rotation = None
        self._anchor = None
        self.layer = layer
        # TODO: add more guid labeles that are relevant to the shape object
        self.settings.update(type(self).SETTINGS)
        if settings:
            self.settings.update(settings)

    @property
    def shape(self):
        """The shape associated with the object."""
        return self._shape

    @shape.setter
    def shape(self, shape):
        self._shape = shape

    @property
    def location(self):
        """:class:`compas.geometry.Point`:
        The location of the object.
        Default is the origin of the world coordinate system.
        The object transformation is applied relative to this location.
        Setting this location will make a copy of the provided point object.
        Moving the original point will thus not affect the object's location.
        """
        if not self._location:
            self._location = Point(0, 0, 0)
        return self._location

    @location.setter
    def location(self, location):
        self._location = Point(*location)

    @property
    def scale(self):
        """float:
        A uniform scaling factor for the object in the scene.
        The scale is applied relative to the location of the object in the scene.
        """
        if not self._scale:
            self._scale = 1.0
        return self._scale

    @scale.setter
    def scale(self, scale):
        self._scale = scale

    @property
    def rotation(self):
        """list of float:
        The rotation angles around the 3 axis of the coordinate system
        with the origin placed at the location of the object in the scene.
        """
        if not self._rotation:
            self._rotation = [0, 0, 0]
        return self._rotation

    @rotation.setter
    def rotation(self, rotation):
        self._rotation = rotation

    @property
    def anchor(self):
        """The vertex of the mesh that is anchored to the location of the object."""
        return self._anchor

    @anchor.setter
    def anchor(self, vertex):
        if self.shape.middle.has_vertex(vertex):
            self._anchor = vertex

    # @property  # See if this property makes sense or it needs to be vertex_xyz_intrados ...
    # def vertex_xyz(self):
    #     """dict : The view coordinates of the mesh object."""
    #     origin = Point(0, 0, 0)
    #     if self.anchor is not None:
    #         xyz = self.mesh.vertex_attributes(self.anchor, 'xyz')
    #         point = Point(* xyz)
    #         T1 = Translation.from_vector(origin - point)
    #         S = Scale.from_factors([self.scale] * 3)
    #         R = Rotation.from_euler_angles(self.rotation)
    #         T2 = Translation.from_vector(self.location)
    #         X = T2 * R * S * T1
    #     else:
    #         S = Scale.from_factors([self.scale] * 3)
    #         R = Rotation.from_euler_angles(self.rotation)
    #         T = Translation.from_vector(self.location)
    #         X = T * R * S
    #     intrados_transformed = self.intrados.transformed(X)
    #     extrados_transformed = self.extrados.transformed(X)
    #     middle_transformed = self.middle.transformed(X)
    #     vertex_xyz = {vertex: mesh.vertex_attributes(vertex, 'xyz') for vertex in mesh.vertices()}
    #     return vertex_xyz

    @property
    def guid_intrados(self):
        """dict: Map between Rhino object GUIDs and intrados identifiers."""
        return self._guid_intrados

    @guid_intrados.setter
    def guid_intrados(self, values):
        self._guid_intrados = dict(values)

    @property
    def guid_extrados(self):
        """dict: Map between Rhino object GUIDs and intrados identifiers."""
        return self._guid_extrados

    @guid_extrados.setter
    def guid_extrados(self, values):
        self._guid_extrados = dict(values)

    @property
    def guid_middle(self):
        """dict: Map between Rhino object GUIDs and intrados identifiers."""
        return self._guid_middle

    @guid_middle.setter
    def guid_middle(self, values):
        self._guid_middle = dict(values)

    @property
    def guids(self):
        """list: The GUIDs of all Rhino objects created by this artist."""
        guids = self._guids
        guids += list(self.guid_intrados.keys())
        guids += list(self.guid_extrados.keys())
        guids += list(self.guid_middle.keys())
        return guids

    def clear(self):
        """Clear all Rhino objects associated with this object.
        """
        compas_rhino.delete_objects(self.guids, purge=True)
        self._guids = []
        self._guid_intrados = {}
        self._guid_extrados = {}
        self._guid_middle = {}

    def draw(self):
        """Draw the shape object.

        The visible components, display properties and visual style of the shape
        drawn by this method can be fully customised using the configuration items
        in the settings dict: ``FormArtist.settings``.

        The method will clear the scene of any objects it has previously drawn
        and keep track of any newly created objects using their GUID.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        self.clear()

        self.artist.shape = self.shape

        if not self.visible:
            return

        # intrados
        if self.settings['show.intrados']:
            guid = self.artist.draw_intrados(color=self.settings['color.intrados'], layer=self.layer + '::Intrados')
            self.guid_intrados = zip(guid, 'intrados')

        # extrados
        if self.settings['show.extrados']:
            guid = self.artist.draw_extrados(color=self.settings['color.extrados'], layer=self.layer + '::Extrados')
            self.guid_extrados = zip(guid, 'extrados')

        # middle
        if self.settings['show.middle']:
            guid = self.artist.draw_middle(color=self.settings['color.middle'], layer=self.layer + '::Middle')
            self.guid_middle = zip(guid, 'middle')

        self.redraw()



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino

from compas_tno.diagrams import FormDiagram


__commandname__ = "TNO_form_from_library"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']

    answer = compas_rhino.rs.GetString("Geometry of the Form Diagram", "Cancel", ["Circular", "Rectangular", "Cancel"])
    if not answer:
        return
    if answer == "Cancel":
        return

    data = {}
    if answer == "Circular":
        center = compas_rhino.rs.GetPoint(message='Select the center of the diagram')
        if not center:
            return
        data['center'] = [center[0], center[1]]

        radius = compas_rhino.rs.GetReal("Set Radius", 5.0)
        if not radius:
            radius = compas_rhino.rs.GetDistance(first_pt=center, second_pt_msg='Select Second Point to find radius')
        if not radius:
            return
        data['radius'] = radius

        layout = compas_rhino.rs.GetString("Set the layout of the diagram (See docs)", "Cancel", ["radial_fd", "radial_spaced_fd", "spiral_fd", "Cancel"])
        if not layout:
            return
        if layout == "Cancel":
            return
        data['type'] = layout

        discr_1 = compas_rhino.rs.GetInteger(message="Radial discretisaton (#hoops)", number=8)
        if not discr_1:
            return
        discr_2 = compas_rhino.rs.GetInteger(message="Meridian discretisaton", number=12)
        if not discr_2:
            return
        data['discretisation'] = [discr_1, discr_2]

        oculus = compas_rhino.rs.GetString("Add Oculus?", "False", ["True", "False", "Cancel"])
        if oculus == "True":
            radius_oc = compas_rhino.rs.GetReal("Radius of oculus (<{0})".format(radius), radius/4)
            if radius_oc:
                data['r_oculus'] = radius_oc

        if layout in ["radial_fd", "radial_spaced_fd"]:
            diagonals = compas_rhino.rs.GetString("Add Diagonals?", "False", ["True", "False", "Cancel"])
            if diagonals == "True":
                data['diagonal'] = True
                diagonals_direction = compas_rhino.rs.GetString("Choose direction", "right", ["right", "left"])
                data['partial_diagonal'] = diagonals_direction

    if answer == "Rectangular":
        rect = compas_rhino.rs.GetRectangle(mode=1, prompt1='Select the first corner of the diagram', prompt2='Select the second corner of the diagram')
        if not rect:
            return
        xy_coords = [[], []]
        for corner in rect:
            xy_coords[0].append(corner[0])
            xy_coords[1].append(corner[1])

        data['xy_span'] = [[min(xy_coords[0]), max(xy_coords[0])], [min(xy_coords[1]), max(xy_coords[1])]]

        layout = compas_rhino.rs.GetString("Set the layout of the diagram (See docs)", "Cancel", ["cross_fd", "fan_fd", "cross_with_diagonal", "ortho", "Cancel"])
        if not layout:
            return
        if layout == "Cancel":
            return
        data['type'] = layout

        discr = compas_rhino.rs.GetInteger(message="Select the discretisation", number=10)
        if not discr:
            return
        data['discretisation'] = discr

        fix = compas_rhino.rs.GetString("Select fixity", "Cancel", ["all", "corners", "Cancel"])
        if not fix:
            return
        if fix == "Cancel":
            return
        data['fix'] = fix

    scene.clear()

    form = FormDiagram.from_library(data)

    objects = scene.find_by_name('Form')
    if not objects:
        scene.add(form, name='Form', layer='TNO::FormDiagram')
    else:
        formobject = objects[0]
        formobject.diagram = form

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



from compas_tna.diagrams import ForceDiagram
from compas.datastructures import mesh_dual


class ForceDiagram(ForceDiagram):
    """The ``ForceDiagram`` represents the equilibrium of the forces in the form diagram.

    Parameters
    ----------


    Attributes
    ----------

    """

    def __init__(self):
        super(ForceDiagram, self).__init__()
        self.attributes.update({
            'name': 'Force'})
        self.update_default_vertex_attributes({
            'is_fixed': False,
            'is_param': False})
        self.update_default_edge_attributes({
            'l': 0.0,
            'target_vector': None,
            'target_length': None})

    # --------------------------------------------------------------------------
    # Constructors
    # --------------------------------------------------------------------------

    @classmethod
    def from_formdiagram(cls, formdiagram):
        """Construct a force diagram from a form diagram.

        Parameters
        ----------
        formdiagram : :class:`~compas_tno.diagrams.ForceDiagram`
            The form diagram.

        Returns
        -------
        :class:`~compas_tno.diagrams.ForceDiagram`
        """
        forcediagram = mesh_dual(formdiagram, cls)
        forcediagram.dual = formdiagram
        formdiagram.dual = forcediagram
        return forcediagram

    # --------------------------------------------------------------------------
    # Convenience functions for retrieving attributes of the force diagram.
    # --------------------------------------------------------------------------

    def xy(self):
        """The XY coordinates of the vertices.

        Returns
        -------
        list
        """
        return self.vertices_attributes('xy')

    def fixed(self):
        """The identifiers of the fixed vertices.

        Returns
        -------
        list
        """
        return list(self.vertices_where({'is_fixed': True}))

    def anchor(self):
        """Get an anchor to the force diagram.

        Returns
        -------
        int
        """
        return next(self.vertices())

    # --------------------------------------------------------------------------
    # Plot
    # --------------------------------------------------------------------------

    def plot(self):
        """Plot a force diagram with a plotter with all the default settings."""
        from compas_plotters import Plotter
        plotter = Plotter(figsize=(8, 8))
        plotter.add(self)
        plotter.show()

    # --------------------------------------------------------------------------
    # Helpers
    # --------------------------------------------------------------------------

    def edges_where_dual(self, conditions, data=False):
        """Get edges for which a certain condition or set of conditions is true for the corresponding edges in the diagram's dual.

        Parameters
        ----------
        conditions : dict
            A set of conditions in the form of key-value pairs.
            The keys should be attribute names. The values can be attribute
            values or ranges of attribute values in the form of min/max pairs.
        data : bool, optional
            Yield the edges and their data attributes.
            Default is ``False``.

        Yields
        ------
        2-tuple
            The next edge as a (u, v) tuple, if ``data=False``.
            The next edge as a ((u, v), data) tuple, if ``data=True``.

        Examples
        --------
        >>>

        """
        for edge in list(self.edges()):
            is_match = True

            dual_edge = self.dual_edge(edge)
            dual_edge_attr = self.dual.edge_attributes(dual_edge)

            for cond_name, cond_value in conditions.items():
                dual_method = getattr(self.dual, cond_name, None)

                if dual_method and callable(dual_method):
                    dual_value = dual_method(dual_edge)
                elif cond_name in dual_edge_attr:
                    dual_value = dual_edge_attr[cond_name]
                else:
                    is_match = False
                    break

                if isinstance(dual_value, list):
                    if cond_value not in dual_value:
                        is_match = False
                        break
                elif isinstance(cond_value, (tuple, list)):
                    minval, maxval = cond_value
                    if dual_value < minval or dual_value > maxval:
                        is_match = False
                        break
                else:
                    if cond_value != dual_value:
                        is_match = False
                        break

            if is_match:
                if data:
                    yield edge, self.edge_attributes(edge)
                else:
                    yield edge

    def dual_edge(self, edge):
        """Find the cooresponding edge in the diagram's dual.

        Parameters
        ----------
        edge : tuple of int
            The edge identifier.

        Returns
        -------
        tuple (int, int) or None
            The identifier of the dual edge if it exists.
        """
        for u, v in self.dual.face_halfedges(edge[0]):
            if self.dual.halfedge[v][u] == edge[1]:
                if self.dual.has_edge((u, v)):
                    return u, v
                return v, u

    def is_dual_edge_external(self, edge):
        """Verify if the corresponding edge in the diagram's dual is marked as "external".

        Parameters
        ----------
        edge : tuple of int
            The edge identifier.

        Returns
        -------
        bool
        """
        return self.dual.edge_attribute(self.dual_edge(edge), 'is_external')

    def is_dual_edge_reaction(self, edge):
        """Verify if the corresponding edge in the diagram's dual is marked as "reaction".

        Parameters
        ----------
        edge : tuple of int
            The edge identifier.

        Returns
        -------
        bool
        """
        return self.dual.edge_attribute(self.dual_edge(edge), 'is_reaction')

    def is_dual_edge_load(self, edge):
        """Verify if the corresponding edge in the diagram's dual is marked as "load".

        Parameters
        ----------
        edge : tuple of int
            The edge identifier.

        Returns
        -------
        bool
        """
        return self.dual.edge_attribute(self.dual_edge(edge), 'is_load')

    def is_dual_edge_ind(self, edge):
        """Verify if the corresponding edge in the diagram's dual is marked as "independent".

        Parameters
        ----------
        edge : tuple of int
            The edge identifier.

        Returns
        -------
        bool
        """
        return self.dual.edge_attribute(self.dual_edge(edge), 'is_ind')

    def dual_edge_force(self, edge):
        """Retrieve the force in the corresponding edge of the diagram's dual.

        Parameters
        ----------
        edge : tuple(int, int)
            The edge identifier.

        Returns
        -------
        float
        """
        return self.dual.edge_attribute(self.dual_edge(edge), 'f')

    def dual_edge_angledeviation(self, edge):
        """Retrieve the angle deviation in the corresponding edge of the diagram's dual.

        Parameters
        ----------
        edge : tuple(int, int)
            The edge identifier.

        Returns
        -------
        float
        """
        return self.dual.edge_attribute(self.dual_edge(edge), 'a')

    def edge_index(self, form=None):
        """Construct a mapping between the identifiers of edges and the corresponding indices in a list of edges.

        Parameters
        ----------
        form : :class:`compas_ags.diagrams.FormDiagram`, optional
            If a form diagra is provided as reference, the list of edges is ordered such that it corresponds
            to the natural ordering of edges in the form diagram.

        Returns
        -------
        dict
            Mapping between edge identifiers and the correpsonding indices of the edges in a list.

        """
        if not form:
            return {edge: index for index, edge in enumerate(self.edges())}
            # return {edge: index for index, edge in enumerate(form.edges_where({'_is_edge': True}))}
        edge_index = dict()
        # for index, (u, v) in enumerate(form.edges()):
        for index, (u, v) in enumerate(form.edges_where({'_is_edge': True})):
            f1 = form.halfedge[u][v]
            f2 = form.halfedge[v][u]
            edge_index[f1, f2] = index
            # the weird side-effect of this is that edges get rotated if necessary
        return edge_index

    def ordered_edges(self, form):
        """"Construct a list of edges with the same order as the corresponding edges of the form diagram.

        Parameters
        ----------
        form : :class:`compas_ags.diagrams.FormDiagram`

        Returns
        -------
        list
        """
        edge_index = self.edge_index(form=form)
        index_edge = {index: edge for edge, index in edge_index.items()}
        edges = [index_edge[index] for index in range(self.number_of_edges())]
        return edges
